# Copyright (c) 2024, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

import os

import cudf
import numpy as np
from dask import dataframe as dd


class JaccardSimilarity:
    def __init__(
        self,
        id_field="id",
        anchor_id_fields=["anchor_0_id", "anchor_1_id"],
        text_field="text",
        ngram_width=5,
    ):
        self.id_field = id_field
        self.anchor_id_fields = anchor_id_fields
        self.text_field = text_field
        self.anchor_id = f"anchor_{id_field}"
        self.left_id = f"{self.id_field}_x"
        self.right_id = f"{self.id_field}_y"
        self.ngram_width = ngram_width

    def __call__(DocumentDataset):
        raise NotImplementedError

    def jaccard_compute(self, shuffled_docs_path):
        paths = [
            entry.path
            for entry in os.scandir(shuffled_docs_path)
            if not entry.path.endswith(".txt")
        ]
        meta_df = cudf.DataFrame(
            {
                self.left_id: ["x"],
                self.right_id: ["y"],
                "jaccard": np.float32([0.0]),
            }
        )
        result_df = dd.from_map(
            self._compute_jaccard_on_1_partition, paths, meta=meta_df
        ).reset_index(drop=True)
        return result_df

    def _compute_jaccard_on_1_partition(self, path):
        try:
            df = cudf.read_parquet(path)
            pair_df = self._compute_jaccard_and_create_pair_df(df)
        except OverflowError:
            paths = [entry.path for entry in os.scandir(os.path.join(path))]
            anchor_df_str_size_ls = [
                self._get_anchor_docs_and_string_size(path) for path in paths
            ]
            anchor_df = cudf.concat(
                [anchor_doc for anchor_doc, _ in anchor_df_str_size_ls],
                ignore_index=True,
            ).drop_duplicates()
            df_str_size = [str_size for _, str_size in anchor_df_str_size_ls]
            paths = JaccardSimilarity._create_bins(
                df_str_size, np.iinfo(np.int32).max // 10
            )
            pair_dfs = []
            for path in paths:
                print(path)
                df = cudf.read_parquet(path).reset_index(drop=True)
                df = cudf.concat([df, anchor_df], ignore_index=True)
                pair_df = self._compute_jaccard_and_create_pair_df(df)
                pair_dfs.append(pair_df)
            pair_df = cudf.concat(pair_dfs, ignore_index=True)
        return pair_df

    def _get_anchor_docs_and_string_size(self, path):
        df = cudf.read_parquet(path)
        str_bytes = df[self.text_field].str.byte_count().sum()
        is_anchor_flag = df[self.id_field] == df[self.anchor_id_fields[0]]
        for anchor_id in self.anchor_id_fields[1:]:
            is_anchor_flag = is_anchor_flag | (df[self.id_field] == df[anchor_id])
        anchor_df = df[is_anchor_flag].reset_index(drop=True)
        return anchor_df, {"path": path, "str_bytes": str_bytes}

    @staticmethod
    def _create_bins(path_dicts, max_size):
        path_dicts.sort(key=lambda x: x["str_bytes"], reverse=True)
        bins, bin_sizes = [], []
        for path_d in path_dicts:
            new_path, new_size = path_d["path"], path_d["str_bytes"]
            for i, bin_size in enumerate(bin_sizes):
                if bin_size + new_size <= max_size:
                    bins[i].append(new_path)
                    bin_sizes[i] += new_size
                    new_size = 0
                    break
            if new_size:
                bins.append([new_path])
                bin_sizes.append(new_size)
        return bins

    def _compute_jaccard_and_create_pair_df(self, df):
        df = df.drop_duplicates(
            subset=[self.id_field] + self.anchor_id_fields, ignore_index=True
        )
        anchor_columns = self.anchor_id_fields
        id_field = self.id_field
        result_ls = []
        try:
            for anchor_col in anchor_columns:
                doc_df = df[[id_field, self.text_field, anchor_col]]
                doc_df = doc_df.rename(columns={anchor_col: self.anchor_id})
                doc_df = doc_df[doc_df[id_field] != doc_df[self.anchor_id]]
                anchor_df = self._get_anchor_df(df, anchor_col)
                result_df = self._compute_jaccard_pair(doc_df, anchor_df)
                result_ls.append(result_df)

            return cudf.concat(result_ls)
        except OverflowError as e:
            print(
                "Failed with  OverflowError in compute_jaccard_and_create_pair_df",
                flush=True,
            )
            print(df, flush=True)
            print("--" * 30)
            print("Error")
            print("---" * 30)
            raise e

    def _get_anchor_df(self, df, anchor_col):
        anchor_df = df[df[self.id_field] == df[anchor_col]]
        anchor_df = anchor_df.reset_index(drop=True)
        anchor_df = anchor_df[[anchor_col, self.text_field]]
        anchor_df = anchor_df.rename(columns={anchor_col: self.anchor_id})
        return anchor_df

    def _compute_jaccard_pair(self, docs_df, anchor_df):
        nrows_at_once = JaccardSimilarity._get_max_num_rows_to_process_once(
            df=docs_df, text_field=self.text_field
        )
        result_ls = []
        for i in range(0, docs_df.shape[0], nrows_at_once):
            pair_df = docs_df[i : i + nrows_at_once]
            pair_df = pair_df.merge(anchor_df, on=self.anchor_id)
            pair_df = pair_df.rename(
                columns={self.id_field: self.left_id, self.anchor_id: self.right_id}
            )
            mask = pair_df[self.left_id] != pair_df[self.right_id]
            pair_df = pair_df[mask].reset_index(drop=True)
            if len(pair_df) == 0:
                result_df = self._create_empty_jaccard_result()
            else:
                result_df = self._compute_jaccard_partition(pair_df)
            result_ls.append(result_df)
        if len(result_ls) == 0:
            return self._create_empty_jaccard_result()
        df_pair = cudf.concat(result_ls)
        return df_pair

    def _create_empty_jaccard_result(self):
        df = cudf.DataFrame()
        df[self.left_id] = "x"
        df[self.right_id] = "y"
        df["jaccard"] = np.empty(shape=0, dtype=np.float32)
        return df

    def _compute_jaccard_partition(self, df):
        text_x = f"{self.text_field}_x"
        text_y = f"{self.text_field}_y"
        df["jaccard"] = df[text_x].str.jaccard_index(df[text_y], width=self.ngram_width)
        df.drop(columns=[text_x, text_y], inplace=True)
        return df

    @staticmethod
    def _get_max_num_rows_to_process_once(df, text_field):
        nbytes = df[text_field].str.byte_count().sum()
        # Number of exmploded bytes
        exploded_bytes = nbytes * 5 * 2
        max_chars_allowed = 2_147_483_647
        byte_ratio = int(exploded_bytes) // max_chars_allowed
        if byte_ratio > 1:
            nrows_at_once = len(df) // byte_ratio
        else:
            nrows_at_once = len(df)

        nrows_at_once = max(1, nrows_at_once)
        return nrows_at_once



# Copyright (c) 2024, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import argparse
import os
import shutil
from functools import partial
from typing import Any

from docbuilder import TedTalksDownloader

from nemo_curator import ParallelScoreFilter, Sequential
from nemo_curator.datasets.parallel_dataset import ParallelDataset
from nemo_curator.filters import (
    HistogramFilter,
    LengthRatioFilter,
    QualityEstimationFilter,
)
from nemo_curator.utils.distributed_utils import get_client
from nemo_curator.utils.script_utils import ArgumentHelper

TED_DE_URL = "https://www.cs.jhu.edu/~kevinduh/a/multitarget-tedtalks/t/en-de/raw/ted_train_en-de.raw.de"
TED_EN_URL = "https://www.cs.jhu.edu/~kevinduh/a/multitarget-tedtalks/t/en-de/raw/ted_train_en-de.raw.en"
SRC_LANG = "en"
TGT_LANG = "de"

SCRIPT_DIR_PATH = os.path.dirname(os.path.abspath(__file__))
DATA_DIR = os.path.join(SCRIPT_DIR_PATH, "data")


def download_files() -> str:
    downloader = TedTalksDownloader(DATA_DIR)

    # ted files is a dict of filenames, 'src', 'tgt' keys and their corresponding file paths as values
    ted_files = downloader.download(TED_EN_URL, TED_DE_URL, force=False)
    return ted_files["src"], ted_files["tgt"]


def filter_dataset(dataset: ParallelDataset, gpu: bool = False) -> ParallelDataset:
    filters = Sequential(
        [
            LengthRatioFilter(
                max_ratio=2,
                src_lang=SRC_LANG,
                tgt_lang=TGT_LANG,
                score_field="length_ratio",
                score_type=float,
            ),
            ParallelScoreFilter(
                HistogramFilter(lang=SRC_LANG),
                HistogramFilter(lang=TGT_LANG),
                src_score="src_hist",
                tgt_score="tgt_hist",
                score_type=int,
            ),
        ]
    )

    if gpu:
        filters.modules.append(
            QualityEstimationFilter(
                "comet-qe",
                cutoff=-0.25,
                gpu=gpu,
                metadata_fields=["src_lang", "tgt_lang"],
                score_type=float,
            )
        )
    else:
        print("Running on CPU, so skipping QE filtering to save time")

    filtered_dataset = filters(dataset)
    return filtered_dataset


def run_curation_pipeline(args: Any, src_file: str, tgt_file: str) -> None:
    # Initialize the Dask cluster.
    client = get_client(**ArgumentHelper.parse_client_args(args))
    print(f"Running curation pipeline on '{src_file} and {tgt_file}'...")

    print("Reading the data...")

    bitext_dataset = ParallelDataset.read_simple_bitext(
        src_file, tgt_file, SRC_LANG, TGT_LANG, add_filename=True, npartitions=16
    )
    curation_steps = Sequential(
        [
            partial(filter_dataset, gpu=(args.device == "gpu")),
        ]
    )

    dataset = curation_steps(bitext_dataset)
    print("Executing the pipeline...")
    dataset = dataset.persist()

    print(f"Original dataset length: {len(bitext_dataset.df)}")
    print(f"After dataprep: {len(dataset.df)}")
    print("Writing the results to disk...")

    # Overwrite existing files in the curated directory.
    out_path = os.path.join(DATA_DIR, "curated")

    if os.path.isdir(out_path):
        shutil.rmtree(out_path)

    os.makedirs(out_path)
    dataset.to_bitext(out_path, write_to_filename=True)
    client.close()


def main():
    parser = argparse.ArgumentParser()
    args = ArgumentHelper(parser).add_distributed_args().parse_args()
    # Limit the total number of workers to ensure we don't run out of memory.
    args.n_workers = min(args.n_workers, 8)

    src_file, tgt_file = download_files()
    run_curation_pipeline(args, src_file, tgt_file)


if __name__ == "__main__":
    main()



# Copyright (c) 2024, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import gzip
import json
import math
import os
import pickle
import random
import string
import tempfile
from datetime import datetime, timedelta
from typing import List, Literal

import pandas as pd
import pytest
from dask import dataframe as dd
from dask.dataframe.utils import assert_eq

from nemo_curator.datasets import DocumentDataset
from nemo_curator.utils.distributed_utils import (
    read_single_partition,
    single_partition_write_with_filename,
    write_to_disk,
)
from nemo_curator.utils.file_utils import get_all_files_paths_under


def _generate_dummy_dataset(num_rows: int = 50) -> str:
    # Function to generate a shuffled sequence of integers
    def shuffled_integers(length: int = num_rows) -> int:
        # Create a list of numbers from 0 to length - 1
        integers = list(range(length))

        # Shuffle the list
        random.shuffle(integers)

        # Yield one number from the list each time the generator is invoked
        for integer in integers:
            yield integer

    # Function to generate a random string of a given length
    def generate_random_string(length: int = 10) -> str:
        characters = string.ascii_letters + string.digits  # Alphanumeric characters

        return "".join(random.choice(characters) for _ in range(length))

    # Function to generate a random datetime
    def generate_random_datetime() -> str:
        # Define start and end dates
        start_date = datetime(1970, 1, 1)  # Unix epoch
        end_date = datetime.now()  # Current date

        # Calculate the total number of seconds between the start and end dates
        delta = end_date - start_date
        total_seconds = int(delta.total_seconds())

        # Generate a random number of seconds within this range
        random_seconds = random.randint(0, total_seconds)

        # Add the random number of seconds to the start date to get a random datetime
        random_datetime = start_date + timedelta(seconds=random_seconds)

        # Convert to UTC and format the datetime
        random_datetime_utc = random_datetime.strftime("%Y-%m-%dT%H:%M:%S.%fZ")

        return random_datetime_utc

    # Generate the corpus
    corpus = []
    for integer in shuffled_integers():
        corpus.append(
            json.dumps(
                {
                    "id": integer,
                    "date": generate_random_datetime(),
                    "text": generate_random_string(random.randint(5, 100)),
                }
            )
        )

    # Return the corpus
    return "\n".join(corpus)


@pytest.fixture
def jsonl_dataset():
    return _generate_dummy_dataset(num_rows=10)


class TestIO:
    def test_meta_dict(self, jsonl_dataset):
        with tempfile.NamedTemporaryFile(suffix=".jsonl") as temp_file:
            # Write the corpus to the file
            temp_file.write(jsonl_dataset.encode("utf-8"))

            # Flush the data to ensure it's written to disk
            temp_file.flush()

            # Move the cursor to the beginning of the file before reading
            temp_file.seek(0)

            # Read the dataset
            dataset = DocumentDataset.read_json(
                temp_file.name, input_meta={"id": float}
            )

        output_meta = str({col: str(dtype) for col, dtype in dataset.df.dtypes.items()})

        expected_meta = "{'id': 'float64'}"

        assert (
            output_meta == expected_meta
        ), f"Expected: {expected_meta}, got: {output_meta}"

    def test_meta_str(self, jsonl_dataset):
        with tempfile.NamedTemporaryFile(suffix=".jsonl") as temp_file:
            # Write the corpus to the file
            temp_file.write(jsonl_dataset.encode("utf-8"))

            # Flush the data to ensure it's written to disk
            temp_file.flush()

            # Move the cursor to the beginning of the file before reading
            temp_file.seek(0)

            # Read the dataset
            dataset = DocumentDataset.read_json(
                temp_file.name, input_meta='{"id": "float"}'
            )

        output_meta = str({col: str(dtype) for col, dtype in dataset.df.dtypes.items()})

        expected_meta = "{'id': 'float64'}"

        assert (
            output_meta == expected_meta
        ), f"Expected: {expected_meta}, got: {output_meta}"

    @pytest.mark.parametrize(
        "backend", ["pandas", pytest.param("cudf", marks=pytest.mark.gpu)]
    )
    def test_read_custom(self, jsonl_dataset: str, backend: Literal["pandas", "cudf"]):
        with tempfile.TemporaryDirectory() as tmp_dir:
            num_lines = jsonl_dataset.count("\n")
            for i, line in enumerate(jsonl_dataset.split("\n")):
                with open(os.path.join(tmp_dir, f"test_{i}.pkl"), "wb") as f:
                    pickle.dump(line, f)

            def read_npy_file(
                files: List[str], backend: Literal["cudf", "pandas"], **kwargs
            ):
                if backend == "cudf":
                    import cudf as df_backend
                else:
                    import pandas as df_backend

                df = df_backend.DataFrame(
                    [{**json.loads(pickle.load(open(file, "rb")))} for file in files],
                )
                return df

            dataset = DocumentDataset.read_custom(
                input_files=tmp_dir,
                file_type="pkl",
                read_func_single_partition=read_npy_file,
                files_per_partition=2,
            )
            assert dataset.df.npartitions == math.ceil(num_lines / 2)
            expected_df = pd.DataFrame(map(json.loads, jsonl_dataset.split("\n")))
            pd.testing.assert_frame_equal(
                dataset.df.to_backend("pandas")
                .compute()
                .sort_values(by="id", ignore_index=True),
                expected_df[["date", "id", "text"]].sort_values(
                    by="id", ignore_index=True
                ),  # because we sort columns by name
            )


class TestWriteWithFilename:
    @pytest.mark.parametrize("keep_filename_column", [True, False])
    @pytest.mark.parametrize("file_ext", ["jsonl", "parquet", "jsonl.gz"])
    @pytest.mark.parametrize("filename_col", ["file_name", "filename"])
    def test_multifile_single_partition(
        self, tmp_path, keep_filename_column, file_ext, filename_col
    ):
        if file_ext == "jsonl.gz":
            compression = "gzip"
            file_type = "jsonl"
        else:
            compression = None
            file_type = file_ext

        df = pd.DataFrame({"a": [1, 2, 3], filename_col: ["file0", "file1", "file1"]})

        single_partition_write_with_filename(
            df=df,
            output_file_dir=tmp_path,
            keep_filename_column=keep_filename_column,
            output_type=file_type,
            filename_col=filename_col,
            compression=compression,
        )
        assert os.path.exists(tmp_path / f"file0.{file_ext}")
        assert os.path.exists(tmp_path / f"file1.{file_ext}")

        if not keep_filename_column:
            df = df.drop(filename_col, axis=1)

        df1 = read_single_partition(
            files=[tmp_path / f"file0.{file_ext}"],
            backend="pandas",
            file_type=file_type,
        )
        assert_eq(df1, df.iloc[0:1], check_index=False)

        df2 = read_single_partition(
            files=[tmp_path / f"file1.{file_ext}"],
            backend="pandas",
            file_type=file_type,
        )
        assert_eq(df2, df.iloc[1:3], check_index=False)

    @pytest.mark.parametrize("keep_filename_column", [True, False])
    @pytest.mark.parametrize("file_ext", ["jsonl", "parquet", "jsonl.gz"])
    def test_singlefile_single_partition(
        self,
        tmp_path,
        keep_filename_column,
        file_ext,
    ):
        if file_ext == "jsonl.gz":
            compression = "gzip"
            file_type = "jsonl"
        else:
            compression = None
            file_type = file_ext

        df = pd.DataFrame({"a": [1, 2, 3], "file_name": ["file2", "file2", "file2"]})

        single_partition_write_with_filename(
            df=df,
            output_file_dir=tmp_path,
            keep_filename_column=keep_filename_column,
            output_type=file_type,
            compression=compression,
        )
        assert len(os.listdir(tmp_path)) == 1
        assert os.path.exists(tmp_path / f"file2.{file_ext}")

        if not keep_filename_column:
            df = df.drop("file_name", axis=1)
        got = read_single_partition(
            files=[tmp_path / f"file2.{file_ext}"],
            backend="pandas",
            file_type=file_type,
        )
        assert_eq(got, df)

    def test_multifile_single_partition_error(self, tmp_path):
        df = pd.DataFrame({"a": [1, 2, 3], "file_name": ["file0", "file1", "file1"]})

        with pytest.raises(ValueError, match="Unknown output type"):
            single_partition_write_with_filename(
                df=df, output_file_dir=tmp_path, output_type="pickle"
            )

    # Test multiple partitions where we need to append to existing files
    @pytest.mark.parametrize(
        "file_ext, read_f",
        [
            ("jsonl", DocumentDataset.read_json),
            ("parquet", DocumentDataset.read_parquet),
            ("jsonl.gz", DocumentDataset.read_json),
        ],
    )
    @pytest.mark.parametrize("filename_col", ["file_name", "filename"])
    def test_multifile_multi_partition(self, tmp_path, file_ext, read_f, filename_col):
        if file_ext == "jsonl.gz":
            compression = "gzip"
            file_type = "jsonl"
        else:
            compression = None
            file_type = file_ext

        df1 = pd.DataFrame({"a": [1, 2, 3], filename_col: ["file1", "file2", "file2"]})
        df2 = df1.copy()
        df2[filename_col] = "file3"
        df3 = df1.copy()
        df3[filename_col] = ["file4", "file5", "file6"]
        ddf = dd.concat([df1, df2, df3])
        ddf[filename_col] = ddf[filename_col] + f".{file_ext}"
        write_to_disk(
            df=ddf,
            output_path=tmp_path / file_ext,
            write_to_filename=filename_col,
            output_type=file_type,
            compression=compression,
        )

        got_df = read_f(
            str(tmp_path / file_ext),
            blocksize=None,
            files_per_partition=2,
            backend="pandas",
            add_filename=filename_col,
        ).df

        assert_eq(got_df, ddf, check_index=False)


class TestFileExtensions:
    def test_keep_extensions(self, tmp_path):
        json_1 = pd.DataFrame({"a": [1, 2, 3], "b": [4, 5, 6]})
        json_1.to_json(tmp_path / "json_1.jsonl", orient="records", lines=True)
        json_2 = pd.DataFrame({"a": [7, 8, 9], "b": [10, 11, 12]})
        json_2.to_json(tmp_path / "json_2.jsonl", orient="records", lines=True)

        json_df = pd.concat([json_1, json_2])

        parquet_file = pd.DataFrame({"a": [13, 14, 15], "b": [16, 17, 18]})
        parquet_file.to_parquet(tmp_path / "parquet_file.parquet")
        csv_file = pd.DataFrame({"a": [19, 20, 21], "b": [22, 23, 24]})
        csv_file.to_csv(tmp_path / "csv_file.csv")

        with pytest.raises(RuntimeError):
            doc = DocumentDataset.read_json(str(tmp_path))

        input_files = get_all_files_paths_under(str(tmp_path), keep_extensions="jsonl")
        input_dataset = DocumentDataset.read_json(input_files)
        assert json_df.equals(input_dataset.df.compute())

        input_files = get_all_files_paths_under(
            str(tmp_path), keep_extensions=["jsonl"]
        )
        input_dataset = DocumentDataset.read_json(input_files)
        assert json_df.equals(input_dataset.df.compute())

        input_files = get_all_files_paths_under(
            str(tmp_path), keep_extensions=["jsonl", "parquet"]
        )
        assert sorted(input_files) == [
            str(tmp_path / "json_1.jsonl"),
            str(tmp_path / "json_2.jsonl"),
            str(tmp_path / "parquet_file.parquet"),
        ]

    def test_write_single_jsonl_file(self, tmp_path):
        json_df = pd.DataFrame({"a": [1, 2, 3], "b": [4, 5, 6]})
        json_df.to_json(tmp_path / "json_file.jsonl", orient="records", lines=True)

        input_path = str(tmp_path / "json_file.jsonl")
        output_path = str(tmp_path / "single_output.jsonl")
        doc = DocumentDataset.read_json(input_path)
        doc.to_json(output_path)

        result = DocumentDataset.read_json(output_path)
        assert json_df.equals(result.df.compute())


class TestPartitionOn:
    def test_partition_on_and_write_to_filename_error(self, tmp_path):
        """Verify that using partition_on and write_to_filename together raises an error."""
        df = pd.DataFrame(
            {
                "id": [1, 2, 3],
                "file_name": ["f1", "f1", "f1"],
                "category": ["A", "B", "A"],
            }
        )
        ddf = dd.from_pandas(df, npartitions=1)
        dataset = DocumentDataset(ddf)
        with pytest.raises(
            ValueError,
            match="Cannot use both partition_on and write_to_filename parameters simultaneously.",
        ):
            dataset.to_json(
                output_path=str(tmp_path / "output"),
                write_to_filename=True,  # Intentionally provided to trigger the error
                partition_on="category",
            )

    @pytest.mark.parametrize(
        "backend", ["pandas", pytest.param("cudf", marks=pytest.mark.gpu)]
    )
    @pytest.mark.parametrize(
        "category_values",
        [
            ["A", "B", "A", "B"],
            [10, 20, 10, 20],
            [1.0, 2.0, 1.0, 2.0],
        ],
    )
    @pytest.mark.parametrize("compression", [None, "gzip"])
    def test_write_to_disk_with_partition_on_jsonl(
        self, tmp_path, backend, category_values, compression
    ):
        """
        Test writing a partitioned JSONL dataset.

        The function is expected to create subdirectories in the output directory
        with names of the form 'category=<value>' for each unique partition column value.
        """
        df = pd.DataFrame(
            {"id": [1, 2, 3, 4], "category": category_values, "value": [10, 20, 30, 40]}
        )
        ddf = dd.from_pandas(df, npartitions=2)
        ddf = ddf.to_backend(backend)
        output_dir = tmp_path / "output_jsonl"
        dataset = DocumentDataset(ddf)
        dataset.to_json(
            output_path=str(output_dir),
            partition_on="category",
            compression=compression,
        )
        # Check that the output directory contains subdirectories for each partition.
        # Unique partition values (as strings) to be used in the directory names.
        unique_partitions = {str(x) for x in category_values}
        for part in unique_partitions:
            expected_dir = output_dir / f"category={part}"
            assert expected_dir.exists(), f"Expected directory {expected_dir} not found"

        # For each partition directory, load the JSONL files and verify that all records have the correct partition value.
        # (Here we assume the files are written with extension ".part")
        for part_dir in output_dir.glob("category=*"):
            # The partition value is taken from the directory name.
            partition_value = part_dir.name.split("=")[-1]
            jsonl_files = list(part_dir.glob("*.part"))
            assert (
                jsonl_files
            ), f"No JSONL files found in partition directory {part_dir}"
            for file in jsonl_files:
                with (
                    gzip.open(file, "rt") if compression == "gzip" else open(file, "r")
                ) as f:
                    for line in f:
                        record = json.loads(line)
                        if "category" in record:
                            # Compare as strings, to work with both integer and string partition values.
                            assert (
                                str(record["category"]) == partition_value
                            ), f"Record partition value {record['category']} does not match directory {partition_value}"

    @pytest.mark.parametrize(
        "backend", ["pandas", pytest.param("cudf", marks=pytest.mark.gpu)]
    )
    @pytest.mark.parametrize(
        "category_values",
        [
            ["A", "B", "A", "B"],
            [10, 20, 10, 20],
            [1.0, 2.0, 1.0, 2.0],
        ],
    )
    def test_write_to_disk_with_partition_on_parquet(
        self, tmp_path, backend, category_values
    ):
        """
        Test writing a partitioned Parquet dataset.

        The test writes a DataFrame partitioned on the 'category' column and then reads it back
        using dd.read_parquet. The output is compared (after sorting) to the original DataFrame.
        """

        df = pd.DataFrame(
            {"id": [1, 2, 3, 4], "category": category_values, "value": [10, 20, 30, 40]}
        )
        ddf = dd.from_pandas(df, npartitions=2)
        ddf = ddf.to_backend(backend)
        output_dir = tmp_path / "output_parquet"
        dataset = DocumentDataset(ddf)
        dataset.to_parquet(output_path=str(output_dir), partition_on="category")

        # Check that the output directory contains subdirectories for each partition.
        # Unique partition values (as strings) to be used in the directory names.
        unique_partitions = {str(x) for x in category_values}
        for part in unique_partitions:
            expected_dir = output_dir / f"category={part}"
            assert expected_dir.exists(), f"Expected directory {expected_dir} not found"

        ddf_loaded = dd.read_parquet(str(output_dir))
        df_loaded = ddf_loaded.compute().reset_index(drop=True)
        df_loaded["category"] = df_loaded["category"].astype(df["category"].dtype)
        # To ensure a fair comparison, sort the dataframes by 'id' and reindex.
        pd.testing.assert_frame_equal(
            df.sort_values("id").reset_index(drop=True),
            df_loaded.sort_values("id").reset_index(drop=True)[df.columns],
            check_dtype=False,
        )



# Copyright (c) 2024, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import os
from dataclasses import dataclass
from typing import List, Optional

os.environ["RAPIDS_NO_INITIALIZE"] = "1"
from crossfit.backend.torch.hf.model import HFModel
from transformers import AutoConfig, AutoTokenizer

from nemo_curator.classifiers.base import (
    DistributedDataClassifier,
    HFDeberta,
    _get_suggest_memory_for_classifier,
    _run_classifier_helper,
)
from nemo_curator.datasets import DocumentDataset

CONTENT_TYPE_IDENTIFIER = "nvidia/content-type-classifier-deberta"


@dataclass
class ContentTypeModelConfig:
    model: str = "microsoft/deberta-v3-base"
    fc_dropout: float = 0.2
    max_len: int = 1024


class ContentTypeModel(HFModel):
    def __init__(
        self,
        config: ContentTypeModelConfig,
        autocast: bool = False,
        max_mem_gb: Optional[int] = None,
    ):
        self.config = config
        self.autocast = autocast
        if max_mem_gb is None:
            max_mem_gb = _get_suggest_memory_for_classifier()

        super().__init__(self.config.model, max_mem_gb=max_mem_gb)

    def load_model(self, device: str = "cuda"):
        model = HFDeberta.from_pretrained(CONTENT_TYPE_IDENTIFIER)
        model.set_autocast(self.autocast)
        model = model.to(device)
        return model.eval()

    def load_tokenizer(self):
        return AutoTokenizer.from_pretrained(CONTENT_TYPE_IDENTIFIER)

    def load_config(self):
        return AutoConfig.from_pretrained(CONTENT_TYPE_IDENTIFIER)


class ContentTypeClassifier(DistributedDataClassifier):
    """
    ContentTypeClassifier is a text classification model designed to categorize documents into one of 11 distinct speech types based on their content.
    It analyzes and understands the nuances of textual information, enabling accurate classification across a diverse range of content types.
    The pretrained model used by this class is called NemoCurator Content Type Classifier DeBERTa.
    It can be found on Hugging Face here: https://huggingface.co/nvidia/content-type-classifier-deberta.
    This classifier is optimized for running on multi-node, multi-GPU setups to enable fast and efficient inference on large datasets.

    Attributes:
        filter_by (list[str], optional): The classes to filter the dataset by.
                                         If None, all classes will be included. Defaults to None.
        batch_size (int): The number of samples per batch for inference. Defaults to 256.
        text_field (str): The field in the dataset that should be classified.
        pred_column (str): The column name where predictions will be stored. Defaults to "content_pred".
        prob_column (str, optional): The column name where prediction probabilities will be stored. Defaults to None.
        max_chars (int): The maximum number of characters in each document to consider for classification. Defaults to 5000.
        device_type (str): The type of device to use for inference, either "cuda" or "cpu". Defaults to "cuda".
        autocast (bool): Whether to use mixed precision for faster inference. Defaults to True.
        max_mem_gb (int, optional): The maximum amount of memory in GB to allocate for the model. If None,
                                      it defaults to the available GPU memory minus 4 GB.

    """

    def __init__(
        self,
        filter_by: Optional[List[str]] = None,
        batch_size: int = 256,
        text_field: str = "text",
        pred_column: str = "content_pred",
        prob_column: Optional[str] = None,
        max_chars: int = 5000,
        device_type: str = "cuda",
        autocast: bool = True,
        max_mem_gb: Optional[int] = None,
    ):
        config = AutoConfig.from_pretrained(CONTENT_TYPE_IDENTIFIER)

        self.text_field = text_field
        self.prob_column = prob_column
        self.labels = list(config.label2id.keys())
        self.labels.sort(key=lambda x: config.label2id[x])
        self.out_dim = len(self.labels)

        model = ContentTypeModel(
            config=ContentTypeModelConfig, autocast=autocast, max_mem_gb=max_mem_gb
        )

        super().__init__(
            model=model,
            labels=self.labels,
            filter_by=filter_by,
            batch_size=batch_size,
            out_dim=self.out_dim,
            pred_column=pred_column,
            max_chars=max_chars,
            device_type=device_type,
            autocast=autocast,
        )

    def _run_classifier(self, dataset: DocumentDataset) -> DocumentDataset:
        print("Starting content type classifier inference", flush=True)
        df = dataset.df
        df = _run_classifier_helper(
            df=df,
            model=self.model,
            labels=self.labels,
            max_chars=self.max_chars,
            batch_size=self.batch_size,
            label_col=self.pred_column,
            text_field=self.text_field,
            prob_col=self.prob_column,
        )
        return DocumentDataset(df)



# Copyright (c) 2025, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os

os.environ["RAPIDS_NO_INITIALIZE"] = "1"
from .aegis import AegisClassifier, InstructionDataGuardClassifier
from .content_type import ContentTypeClassifier
from .domain import DomainClassifier, MultilingualDomainClassifier
from .fineweb_edu import (
    FineWebEduClassifier,
    FineWebMixtralEduClassifier,
    FineWebNemotronEduClassifier,
)
from .prompt_task_complexity import PromptTaskComplexityClassifier
from .quality import QualityClassifier

__all__ = [
    "DomainClassifier",
    "MultilingualDomainClassifier",
    "QualityClassifier",
    "AegisClassifier",
    "InstructionDataGuardClassifier",
    "FineWebEduClassifier",
    "FineWebMixtralEduClassifier",
    "FineWebNemotronEduClassifier",
    "ContentTypeClassifier",
    "PromptTaskComplexityClassifier",
]



# Copyright (c) 2025, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import argparse

from nemo_curator.utils.script_utils import ArgumentHelper


class TestArgumentHelper:
    def test_argument_helper(self):
        parser = argparse.ArgumentParser()
        argument_helper = ArgumentHelper(parser)
        assert "--version" in argument_helper.parser.format_help()

    def test_attach_bool_arg(self):
        parser = argparse.ArgumentParser()
        argument_helper = ArgumentHelper(parser)
        argument_helper.attach_bool_arg(
            argument_helper.parser, "test", default=False, help="test help"
        )
        assert "--test" in parser.format_help()
        assert "test help" in parser.format_help()
        assert "--no-test" in parser.format_help()

    def test_add_arg_batch_size(self):
        parser = argparse.ArgumentParser()
        argument_helper = ArgumentHelper(parser)
        argument_helper.add_arg_batch_size()
        assert "--batch-size" in argument_helper.parser.format_help()

    def test_add_arg_device(self):
        parser = argparse.ArgumentParser()
        argument_helper = ArgumentHelper(parser)
        argument_helper.add_arg_device()
        assert "--device" in argument_helper.parser.format_help()

    def test_add_arg_enable_spilling(self):
        parser = argparse.ArgumentParser()
        argument_helper = ArgumentHelper(parser)
        argument_helper.add_arg_enable_spilling()
        assert "--enable-spilling" in argument_helper.parser.format_help()

    def test_add_arg_language(self):
        parser = argparse.ArgumentParser()
        argument_helper = ArgumentHelper(parser)
        help = "The language of the dataset to be read in."
        argument_helper.add_arg_language(help=help)
        assert "--language" in argument_helper.parser.format_help()
        assert help in argument_helper.parser.format_help()

    def test_add_arg_log_dir(self):
        parser = argparse.ArgumentParser()
        argument_helper = ArgumentHelper(parser)
        default = "./log"
        argument_helper.add_arg_log_dir(default=default)
        assert "--log-dir" in argument_helper.parser.format_help()

    def test_add_arg_input_data_dir(self):
        parser = argparse.ArgumentParser()
        argument_helper = ArgumentHelper(parser)
        argument_helper.add_arg_input_data_dir()
        assert "--input-data-dir" in argument_helper.parser.format_help()

    def test_add_arg_input_file_type(self):
        parser = argparse.ArgumentParser()
        argument_helper = ArgumentHelper(parser)
        choices = ["jsonl", "pickle", "parquet"]
        argument_helper.add_arg_input_file_type(choices=choices)
        assert "--input-file-type" in argument_helper.parser.format_help()
        assert "jsonl" in argument_helper.parser.format_help()
        assert "pickle" in argument_helper.parser.format_help()
        assert "parquet" in argument_helper.parser.format_help()

    def test_add_arg_input_file_extension(self):
        parser = argparse.ArgumentParser()
        argument_helper = ArgumentHelper(parser)
        argument_helper.add_arg_input_file_extension()
        assert "--input-file-extension" in argument_helper.parser.format_help()

    def test_add_arg_input_local_data_dir(self):
        parser = argparse.ArgumentParser()
        argument_helper = ArgumentHelper(parser)
        argument_helper.add_arg_input_local_data_dir()
        assert "--input-local-data-dir" in argument_helper.parser.format_help()

    def test_add_arg_input_meta(self):
        parser = argparse.ArgumentParser()
        argument_helper = ArgumentHelper(parser)
        argument_helper.add_arg_input_meta()
        assert "--input-meta" in argument_helper.parser.format_help()

    def test_add_arg_input_text_field(self):
        parser = argparse.ArgumentParser()
        argument_helper = ArgumentHelper(parser)
        argument_helper.add_arg_input_text_field()
        assert "--input-text-field" in argument_helper.parser.format_help()

    def test_add_arg_id_column(self):
        parser = argparse.ArgumentParser()
        argument_helper = ArgumentHelper(parser)
        argument_helper.add_arg_id_column()
        assert "--id-column" in argument_helper.parser.format_help()

    def test_add_arg_id_column_type(self):
        parser = argparse.ArgumentParser()
        argument_helper = ArgumentHelper(parser)
        argument_helper.add_arg_id_column_type()
        assert "--id-column-type" in argument_helper.parser.format_help()

    def test_add_arg_minhash_length(self):
        parser = argparse.ArgumentParser()
        argument_helper = ArgumentHelper(parser)
        argument_helper.add_arg_minhash_length()
        assert "--minhash-length" in argument_helper.parser.format_help()

    def test_add_arg_nvlink_only(self):
        parser = argparse.ArgumentParser()
        argument_helper = ArgumentHelper(parser)
        argument_helper.add_arg_nvlink_only()
        assert "--nvlink-only" in argument_helper.parser.format_help()

    def test_add_arg_output_data_dir(self):
        parser = argparse.ArgumentParser()
        argument_helper = ArgumentHelper(parser)
        help = "Output data directory."
        argument_helper.add_arg_output_data_dir(help=help)
        assert "--output-data-dir" in argument_helper.parser.format_help()
        assert help in argument_helper.parser.format_help()

    def test_add_arg_output_dir(self):
        parser = argparse.ArgumentParser()
        argument_helper = ArgumentHelper(parser)
        argument_helper.add_arg_output_dir()
        assert "--output-dir" in argument_helper.parser.format_help()

    def test_add_arg_output_file_type(self):
        parser = argparse.ArgumentParser()
        argument_helper = ArgumentHelper(parser)
        choices = ["jsonl", "pickle", "parquet"]
        argument_helper.add_arg_output_file_type(choices=choices)
        assert "--output-file-type" in argument_helper.parser.format_help()
        assert "jsonl" in argument_helper.parser.format_help()
        assert "pickle" in argument_helper.parser.format_help()
        assert "parquet" in argument_helper.parser.format_help()

    def test_add_arg_output_train_file(self):
        parser = argparse.ArgumentParser()
        argument_helper = ArgumentHelper(parser)
        help = "The output train file."
        argument_helper.add_arg_output_train_file(help=help)
        assert "--output-train-file" in argument_helper.parser.format_help()
        assert help in argument_helper.parser.format_help()

    def test_add_arg_protocol(self):
        parser = argparse.ArgumentParser()
        argument_helper = ArgumentHelper(parser)
        argument_helper.add_arg_protocol()
        assert "--protocol" in argument_helper.parser.format_help()

    def test_add_arg_rmm_pool_size(self):
        parser = argparse.ArgumentParser()
        argument_helper = ArgumentHelper(parser)
        argument_helper.add_arg_rmm_pool_size()
        assert "--rmm-pool-size" in argument_helper.parser.format_help()

    def test_add_arg_scheduler_address(self):
        parser = argparse.ArgumentParser()
        argument_helper = ArgumentHelper(parser)
        argument_helper.add_arg_scheduler_address()
        assert "--scheduler-address" in argument_helper.parser.format_help()

    def test_add_arg_scheduler_file(self):
        parser = argparse.ArgumentParser()
        argument_helper = ArgumentHelper(parser)
        argument_helper.add_arg_scheduler_file()
        assert "--scheduler-file" in argument_helper.parser.format_help()

    def test_add_arg_seed(self):
        parser = argparse.ArgumentParser()
        argument_helper = ArgumentHelper(parser)
        argument_helper.add_arg_seed()
        assert "--seed" in argument_helper.parser.format_help()

    def test_add_arg_set_torch_to_use_rmm(self):
        parser = argparse.ArgumentParser()
        argument_helper = ArgumentHelper(parser)
        argument_helper.add_arg_set_torch_to_use_rmm()
        assert "--set-torch-to-use-rmm" in argument_helper.parser.format_help()

    def test_add_arg_shuffle(self):
        parser = argparse.ArgumentParser()
        argument_helper = ArgumentHelper(parser)
        help = "Shuffle argument help"
        argument_helper.add_arg_shuffle(help=help)
        assert "--shuffle" in argument_helper.parser.format_help()
        assert help in argument_helper.parser.format_help()

    def test_add_arg_text_ddf_blocksize(self):
        parser = argparse.ArgumentParser()
        argument_helper = ArgumentHelper(parser)
        argument_helper.add_arg_text_ddf_blocksize()
        assert "--text-ddf-blocksize" in argument_helper.parser.format_help()

    def test_add_arg_model_path(self):
        parser = argparse.ArgumentParser()
        argument_helper = ArgumentHelper(parser)
        argument_helper.add_arg_model_path()
        assert "--pretrained-model-name-or-path" in argument_helper.parser.format_help()

    def test_add_arg_max_mem_gb_classifier(self):
        parser = argparse.ArgumentParser()
        argument_helper = ArgumentHelper(parser)
        argument_helper.add_arg_max_mem_gb_classifier()
        assert "--max-mem-gb-classifier" in argument_helper.parser.format_help()

    def test_add_arg_autocast(self):
        parser = argparse.ArgumentParser()
        argument_helper = ArgumentHelper(parser)
        argument_helper.add_arg_autocast()
        assert "--autocast" in argument_helper.parser.format_help()

    def test_add_arg_max_chars(self):
        parser = argparse.ArgumentParser()
        argument_helper = ArgumentHelper(parser)
        argument_helper.add_arg_max_chars()
        assert "--max-chars" in argument_helper.parser.format_help()

    def test_distributed_args(self):
        parser = argparse.ArgumentParser()
        argument_helper = ArgumentHelper(parser)
        argument_helper.add_distributed_args()

        assert "--device" in argument_helper.parser.format_help()
        assert "--files-per-partition" in argument_helper.parser.format_help()
        assert "--n-workers" in argument_helper.parser.format_help()
        assert "--num-files" in argument_helper.parser.format_help()
        assert "--nvlink-only" in argument_helper.parser.format_help()
        assert "--protocol" in argument_helper.parser.format_help()
        assert "--rmm-pool-size" in argument_helper.parser.format_help()
        assert "--scheduler-address" in argument_helper.parser.format_help()
        assert "--scheduler-file" in argument_helper.parser.format_help()
        assert "--threads-per-worker" in argument_helper.parser.format_help()

    def test_set_default_n_workers(self):
        parser = argparse.ArgumentParser()
        argument_helper = ArgumentHelper(parser)
        max_mem_gb_per_worker = 10.0
        argument_helper.set_default_n_workers(max_mem_gb_per_worker)
        assert "n_workers" in argument_helper.parser._defaults

    def test_parse_client_args(self):
        parser = argparse.ArgumentParser()
        argument_helper = ArgumentHelper(parser)
        args = argparse.Namespace(device="gpu", n_workers=10, random_arg="abc")
        parsed_args = argument_helper.parse_client_args(args)
        assert parsed_args["cluster_type"] == "gpu"
        assert parsed_args["n_workers"] == 10
        assert "random_arg" not in parsed_args

    def test_parse_distributed_classifier_args(self):
        parser = argparse.ArgumentParser()
        argument_helper = ArgumentHelper(parser)
        parser = argument_helper.parse_distributed_classifier_args()

        assert "--device" in parser.format_help()
        assert "--files-per-partition" in parser.format_help()
        assert "--n-workers" in parser.format_help()
        assert "--num-files" in parser.format_help()
        assert "--nvlink-only" in parser.format_help()
        assert "--protocol" in parser.format_help()
        assert "--rmm-pool-size" in parser.format_help()
        assert "--scheduler-address" in parser.format_help()
        assert "--scheduler-file" in parser.format_help()
        assert "--threads-per-worker" in parser.format_help()
        assert "--enable-spilling" in parser.format_help()
        assert "--set-torch-to-use-rmm" in parser.format_help()
        assert "--max-mem-gb-classifier" in parser.format_help()
        assert "rmm_pool_size" in parser._defaults
        assert "set_torch_to_use_rmm" in parser._defaults

        assert "--input-data-dir" in parser.format_help()
        assert "--output-data-dir" in parser.format_help()
        assert "--input-file-type" in parser.format_help()
        assert "--input-file-extension" in parser.format_help()
        assert "--output-file-type" in parser.format_help()
        assert "--input-text-field" in parser.format_help()
        assert "--batch-size" in parser.format_help()
        assert "--pretrained-model-name-or-path" in parser.format_help()
        assert "--autocast" in parser.format_help()
        assert "--max-chars" in parser.format_help()

    def test_add_distributed_classifier_cluster_args(self):
        parser = argparse.ArgumentParser()
        argument_helper = ArgumentHelper(parser)
        argument_helper.add_distributed_classifier_cluster_args()

        assert "--device" in argument_helper.parser.format_help()
        assert "--files-per-partition" in argument_helper.parser.format_help()
        assert "--n-workers" in argument_helper.parser.format_help()
        assert "--num-files" in argument_helper.parser.format_help()
        assert "--nvlink-only" in argument_helper.parser.format_help()
        assert "--protocol" in argument_helper.parser.format_help()
        assert "--rmm-pool-size" in argument_helper.parser.format_help()
        assert "--scheduler-address" in argument_helper.parser.format_help()
        assert "--scheduler-file" in argument_helper.parser.format_help()
        assert "--threads-per-worker" in argument_helper.parser.format_help()
        assert "--enable-spilling" in argument_helper.parser.format_help()
        assert "--set-torch-to-use-rmm" in argument_helper.parser.format_help()
        assert "--max-mem-gb-classifier" in argument_helper.parser.format_help()
        assert "rmm_pool_size" in argument_helper.parser._defaults
        assert "set_torch_to_use_rmm" in argument_helper.parser._defaults

    def test_parse_gpu_dedup_args(self):
        parser = argparse.ArgumentParser()
        argument_helper = ArgumentHelper(parser)
        argument_helper.parse_gpu_dedup_args()

        assert "--device" in argument_helper.parser.format_help()
        assert "--files-per-partition" in argument_helper.parser.format_help()
        assert "--n-workers" in argument_helper.parser.format_help()
        assert "--num-files" in argument_helper.parser.format_help()
        assert "--nvlink-only" in argument_helper.parser.format_help()
        assert "--protocol" in argument_helper.parser.format_help()
        assert "--rmm-pool-size" in argument_helper.parser.format_help()
        assert "--scheduler-address" in argument_helper.parser.format_help()
        assert "--scheduler-file" in argument_helper.parser.format_help()
        assert "--threads-per-worker" in argument_helper.parser.format_help()

        assert "device" in argument_helper.parser._defaults
        assert "set_torch_to_use_rmm" in argument_helper.parser._defaults

        assert "--input-data-dirs" in argument_helper.parser.format_help()
        assert "--input-json-text-field" in argument_helper.parser.format_help()
        assert "--input-json-id-field" in argument_helper.parser.format_help()
        assert "--log-dir" in argument_helper.parser.format_help()
        assert "--profile-path" in argument_helper.parser.format_help()

    def test_parse_semdedup_args(self):
        parser = argparse.ArgumentParser()
        argument_helper = ArgumentHelper(parser)
        parser = argument_helper.parse_semdedup_args()

        assert "--device" in parser.format_help()
        assert "--files-per-partition" in parser.format_help()
        assert "--n-workers" in parser.format_help()
        assert "--num-files" in parser.format_help()
        assert "--nvlink-only" in parser.format_help()
        assert "--protocol" in parser.format_help()
        assert "--rmm-pool-size" in parser.format_help()
        assert "--scheduler-address" in parser.format_help()
        assert "--scheduler-file" in parser.format_help()
        assert "--threads-per-worker" in parser.format_help()

        assert "--input-data-dir" in parser.format_help()
        assert "--input-file-extension" in parser.format_help()
        assert "--input-file-type" in parser.format_help()
        assert "--input-text-field" in parser.format_help()
        assert "--id-column" in parser.format_help()

        assert "--config-file" in parser.format_help()

        assert "rmm_pool_size" in parser._defaults
        assert "device" in parser._defaults
        assert "set_torch_to_use_rmm" in parser._defaults



# Copyright (c) 2025, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import os
from typing import Optional

os.environ["RAPIDS_NO_INITIALIZE"] = "1"
import torch
from crossfit import op
from crossfit.backend.torch.hf.model import HFModel
from transformers import AutoConfig, AutoModelForSequenceClassification, AutoTokenizer

from nemo_curator.classifiers.base import (
    DistributedDataClassifier,
    _get_suggest_memory_for_classifier,
)
from nemo_curator.datasets import DocumentDataset

FINEWEB_EDU_IDENTIFIER = "HuggingFaceFW/fineweb-edu-classifier"
FINEWEB_MIXTRAL_IDENTIFIER = "nvidia/nemocurator-fineweb-mixtral-edu-classifier"
FINEWEB_NEMOTRON_IDENTIFIER = "nvidia/nemocurator-fineweb-nemotron-4-edu-classifier"


class FinewebEduModel(HFModel):
    def __init__(
        self,
        path_or_name: str,
        max_mem_gb: Optional[int] = None,
        autocast: bool = False,
    ):
        self.path_or_name = path_or_name
        self.autocast = autocast
        if max_mem_gb is None:
            max_mem_gb = _get_suggest_memory_for_classifier()
        super().__init__(path_or_name=path_or_name, max_mem_gb=max_mem_gb)

    def load_model(self, device: str = "cuda"):
        model = AutoModelForSequenceClassification.from_pretrained(self.path_or_name)
        model = model.to(device)
        model = self.configure_forward(model, self.autocast)
        return model

    @staticmethod
    def configure_forward(model, autocast: bool = True):
        original_forward = model.forward

        def custom_forward(*args, **kwargs):
            if autocast:
                with torch.autocast(device_type="cuda"):
                    output = original_forward(*args, **kwargs)
            else:
                output = original_forward(*args, **kwargs)
            return output.logits.squeeze(-1).float()

        model.forward = custom_forward
        return model

    def load_tokenizer(self):
        return AutoTokenizer.from_pretrained(self.path_or_name)

    def load_config(self):
        return AutoConfig.from_pretrained(self.path_or_name)


class _FineWebBaseClassifier(DistributedDataClassifier):
    """
    Parent class for FineWebEduClassifier, FineWebMixtralEduClassifier, and FineWebNemotronEduClassifier,
    since their implementations are almost identical.
    """

    def __init__(
        self,
        fineweb_identifier: str,
        pred_column: str,
        int_column: str,
        quality_label_column: Optional[str],
        batch_size: int = 1024,
        text_field: str = "text",
        max_chars: int = -1,
        device_type: str = "cuda",
        autocast: bool = True,
        max_mem_gb: Optional[int] = None,
    ):
        self.fineweb_identifier = fineweb_identifier

        model = FinewebEduModel(
            path_or_name=fineweb_identifier,
            autocast=autocast,
            max_mem_gb=max_mem_gb,
        )

        self.text_field = text_field
        self.int_column = int_column
        self.quality_label_column = quality_label_column

        super().__init__(
            model=model,
            filter_by=None,  # No filtering as its a numeric score
            batch_size=batch_size,
            pred_column=pred_column,
            max_chars=max_chars,
            device_type=device_type,
            autocast=autocast,
            labels=None,
            out_dim=1,
        )

    def _run_classifier(self, dataset: DocumentDataset) -> DocumentDataset:
        if self.fineweb_identifier == FINEWEB_EDU_IDENTIFIER:
            print("Starting FineWeb-Edu Classifier inference", flush=True)
        elif self.fineweb_identifier == FINEWEB_MIXTRAL_IDENTIFIER:
            print("Starting FineWeb Mixtral Edu Classifier inference", flush=True)
        elif self.fineweb_identifier == FINEWEB_NEMOTRON_IDENTIFIER:
            print("Starting FineWeb Nemotron-4 Edu Classifier inference", flush=True)

        ddf = dataset.df

        pipe = op.Sequential(
            op.Tokenizer(
                self.model,
                cols=[self.text_field],
                tokenizer_type="default",
                max_length=self.model.max_seq_length(),
            ),
            op.Predictor(
                self.model,
                sorted_data_loader=True,
                batch_size=self.batch_size,
                pred_output_col=self.pred_column,
            ),
            keep_cols=ddf.columns.tolist(),
        )
        ddf = pipe(ddf)

        ddf[self.pred_column] = ddf[self.pred_column].where(
            ddf[self.pred_column] >= 0, 0
        )
        ddf[self.pred_column] = ddf[self.pred_column].where(
            ddf[self.pred_column] <= 5, 5
        )
        ddf[self.int_column] = ddf[self.pred_column].round().astype(int)

        if self.quality_label_column is not None:
            ddf[self.quality_label_column] = "high_quality"
            # If the score is less than 2.5, label it as low quality
            ddf[self.quality_label_column] = ddf[self.quality_label_column].mask(
                ddf[self.pred_column] < 2.5, "low_quality"
            )

        return DocumentDataset(ddf)


class FineWebEduClassifier(_FineWebBaseClassifier):
    """
    FineWebEduClassifier is a specialized classifier designed for educational content assessment,
    utilizing the Hugging Face FineWeb EDU Classifier model (https://huggingface.co/HuggingFaceFW/fineweb-edu-classifier).
    This classifier is optimized for running on multi-node, multi-GPU setups to enable fast and efficient inference on large text datasets.

    Attributes:
        batch_size (int): The number of samples per batch for inference. Defaults to 256.
        text_field (str): The column name containing the text data to be classified. Defaults to "text".
        pred_column (str): The column name where prediction scores will be stored. Defaults to "fineweb-edu-score".
        int_column (str): The column name where integer-rounded prediction scores will be stored. Defaults to "fineweb-edu-score-int".
        max_chars (int): The maximum number of characters in each document to consider for classification. If -1, the entire document is considered. Defaults to -1.
        device_type (str): The type of device to use for inference, either "cuda" or "cpu". Defaults to "cuda".
        autocast (bool): Whether to use mixed precision for faster inference. Defaults to True.
        max_mem_gb (int, optional): The maximum amount of memory in GB to allocate for the model. If None,
                                      it defaults to the available GPU memory minus 4 GB.

    """

    def __init__(
        self,
        batch_size: int = 256,
        text_field: str = "text",
        pred_column: str = "fineweb-edu-score",
        int_column="fineweb-edu-score-int",
        max_chars: int = -1,
        device_type: str = "cuda",
        autocast: bool = True,
        max_mem_gb: Optional[int] = None,
    ):
        super().__init__(
            fineweb_identifier=FINEWEB_EDU_IDENTIFIER,
            batch_size=batch_size,
            text_field=text_field,
            pred_column=pred_column,
            int_column=int_column,
            quality_label_column=None,
            max_chars=max_chars,
            device_type=device_type,
            autocast=autocast,
            max_mem_gb=max_mem_gb,
        )


class FineWebMixtralEduClassifier(_FineWebBaseClassifier):
    """
    FineWebMixtralEduClassifier is a specialized classifier designed for educational content assessment,
    utilizing the NemoCurator FineWeb Mixtral Edu Classifier model (https://huggingface.co/nvidia/nemocurator-fineweb-mixtral-edu-classifier).
    It is similar to the FineWeb-Edu classifier and was trained on the same text samples, but using annotations from Mixtral 8x22B-Instruct.
    This classifier is optimized for running on multi-node, multi-GPU setups to enable fast and efficient inference on large text datasets.

    Attributes:
        batch_size (int): The number of samples per batch for inference. Defaults to 256.
        text_field (str): The column name containing the text data to be classified. Defaults to "text".
        pred_column (str): The column name where prediction scores will be stored. Defaults to "fineweb-mixtral-edu-score".
        int_column (str): The column name where integer-rounded prediction scores will be stored. Defaults to "fineweb-mixtral-edu-score-int".
        quality_label_column (str): The column name where a score of >= 2.5 is labeled "high_quality" and otherwise labeled "low_quality". Defaults to "fineweb-mixtral-edu-score-label".
        max_chars (int): The maximum number of characters in each document to consider for classification. If -1, the entire document is considered. Defaults to -1.
        device_type (str): The type of device to use for inference, either "cuda" or "cpu". Defaults to "cuda".
        autocast (bool): Whether to use mixed precision for faster inference. Defaults to True.
        max_mem_gb (int, optional): The maximum amount of memory in GB to allocate for the model. If None,
                                      it defaults to the available GPU memory minus 4 GB.

    """

    def __init__(
        self,
        batch_size: int = 1024,
        text_field: str = "text",
        pred_column: str = "fineweb-mixtral-edu-score",
        int_column: str = "fineweb-mixtral-edu-score-int",
        quality_label_column: str = "fineweb-mixtral-edu-score-label",
        max_chars: int = -1,
        device_type: str = "cuda",
        autocast: bool = True,
        max_mem_gb: Optional[int] = None,
    ):
        super().__init__(
            fineweb_identifier=FINEWEB_MIXTRAL_IDENTIFIER,
            batch_size=batch_size,
            text_field=text_field,
            pred_column=pred_column,
            int_column=int_column,
            quality_label_column=quality_label_column,
            max_chars=max_chars,
            device_type=device_type,
            autocast=autocast,
            max_mem_gb=max_mem_gb,
        )


class FineWebNemotronEduClassifier(_FineWebBaseClassifier):
    """
    FineWebNemotronEduClassifier is a specialized classifier designed for educational content assessment,
    utilizing the NemoCurator FineWeb Nemotron-4 Edu Classifier model (https://huggingface.co/nvidia/nemocurator-fineweb-nemotron-4-edu-classifier).
    It is similar to the FineWeb-Edu classifier and was trained on the same text samples, but using annotations from Nemotron-4-340B-Instruct.
    This classifier is optimized for running on multi-node, multi-GPU setups to enable fast and efficient inference on large text datasets.

    Attributes:
        batch_size (int): The number of samples per batch for inference. Defaults to 256.
        text_field (str): The column name containing the text data to be classified. Defaults to "text".
        pred_column (str): The column name where prediction scores will be stored. Defaults to "fineweb-nemotron-edu-score".
        int_column (str): The column name where integer-rounded prediction scores will be stored. Defaults to "fineweb-nemotron-edu-score-int".
        quality_label_column (str): The column name where a score of >= 2.5 is labeled "high_quality" and otherwise labeled "low_quality". Defaults to "fineweb-nemotron-edu-score-label".
        max_chars (int): The maximum number of characters in each document to consider for classification. If -1, the entire document is considered. Defaults to -1.
        device_type (str): The type of device to use for inference, either "cuda" or "cpu". Defaults to "cuda".
        autocast (bool): Whether to use mixed precision for faster inference. Defaults to True.
        max_mem_gb (int, optional): The maximum amount of memory in GB to allocate for the model. If None,
                                      it defaults to the available GPU memory minus 4 GB.

    """

    def __init__(
        self,
        batch_size: int = 1024,
        text_field: str = "text",
        pred_column: str = "fineweb-nemotron-edu-score",
        int_column: str = "fineweb-nemotron-edu-score-int",
        quality_label_column: str = "fineweb-nemotron-edu-score-label",
        max_chars: int = -1,
        device_type: str = "cuda",
        autocast: bool = True,
        max_mem_gb: Optional[int] = None,
    ):
        super().__init__(
            fineweb_identifier=FINEWEB_NEMOTRON_IDENTIFIER,
            batch_size=batch_size,
            text_field=text_field,
            pred_column=pred_column,
            int_column=int_column,
            quality_label_column=quality_label_column,
            max_chars=max_chars,
            device_type=device_type,
            autocast=autocast,
            max_mem_gb=max_mem_gb,
        )



# Copyright (c) 2024, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import importlib
from abc import ABC, abstractmethod
from typing import Any, Literal, Union

from nemo_curator.filters.bitext_filter import BitextFilter


class DocumentFilter(ABC):
    """
    An abstract base class for text-based document filters.

    This class serves as a template for creating specific document filters
    in the library. Subclasses should implement the abstract methods to
    define custom filtering behavior.
    """

    def __init__(self):
        super().__init__()
        self._name = self.__class__.__name__
        self._sentences = None
        self._paragraphs = None
        self._ngrams = None

    @abstractmethod
    def score_document(self, text: str) -> Any:
        """
        Calculate a score for the given document text.

        This method should be implemented by subclasses to define how
        a document's text is evaluated and scored.

        Args:
            text (str): The text content of the document to be scored.

        Returns:
            Any: A score or set of scores representing the document's
            relevance or quality. The type and structure of the
            return value should be consistent for each subclass.

        Raises:
            NotImplementedError: If the method is not implemented in a subclass.
        """
        raise NotImplementedError(
            "score_document method must be implemented by subclasses"
        )

    @abstractmethod
    def keep_document(self, scores: Any) -> bool:
        """
        Determine whether to keep a document based on its scores.

        This method should be implemented by subclasses to define the
        criteria for keeping or discarding a document based on the
        scores calculated by score_document().

        Args:
            scores (Any): The score or set of scores returned by score_document().
                          The type should match what is returned by score_document().

        Returns:
            bool: True if the document should be kept, False otherwise.

        Raises:
            NotImplementedError: If the method is not implemented in a subclass.
        """
        raise NotImplementedError(
            "keep_document method must be implemented by subclasses"
        )

    @property
    def backend(self) -> Literal["pandas", "cudf", "any"]:
        """
        The dataframe backend the filter operates on.
        Can be 'pandas', 'cudf', or 'any'. Defaults to 'pandas'.
        Returns:
            str: A string representing the dataframe backend the filter needs as input
        """
        return "pandas"

    @property
    def name(self):
        return self._name

    @property
    def sentences(self):
        return self._sentences

    @sentences.setter
    def sentences(self, sentences):
        self._sentences = sentences

    @property
    def paragraphs(self):
        return self._paragraphs

    @paragraphs.setter
    def paragraphs(self, paragraphs):
        self._paragraphs = paragraphs

    @property
    def ngrams(self):
        return self._ngrams

    @ngrams.setter
    def ngrams(self, ngrams):
        self._ngrams = ngrams


def import_filter(filter_path: str) -> Union[DocumentFilter, BitextFilter]:
    """
    Imports a filter under nemo_curator.filters given the module path

    Args:
        filter_path (str): The path to the filter in the form of "nemo_curator.filters.filter_name"

    Returns:
        DocumentFilter: The filter that is at the given path

    Raises:
        ValueError: If the filter_path does not point to a DocumentFilter
    """
    module_path, filter_name = filter_path.rsplit(".", 1)
    filter_module = importlib.import_module(module_path)
    filter_class = getattr(filter_module, filter_name)
    if not issubclass(filter_class, DocumentFilter) and not issubclass(
        filter_class, BitextFilter
    ):
        raise ValueError(
            f"Input filter {filter_class.__name__} must be derived "
            "from DocumentFilter defined in nemo_curator.filters.doc_filter or"
            "from BitextFilter defined in nemo_curator.filters.bitext_filter"
        )
    return filter_class



# Copyright (c) 2024, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import argparse
import importlib
import os
import shutil
import time
from typing import Any, List

import numpy as np
from dask.diagnostics import ProgressBar
from dask.distributed import progress
from retriever_evalset_generator import RetrieverEvalSetGenerator

from nemo_curator import AsyncOpenAIClient, ScoreFilter, Sequential, get_client
from nemo_curator.datasets import DocumentDataset
from nemo_curator.filters import (
    AnswerabilityFilter,
    EasinessFilter,
    NonAlphaNumericFilter,
)
from nemo_curator.modules.filter import Score, ScoreFilter
from nemo_curator.utils.file_utils import get_all_files_paths_under

config = importlib.import_module(
    "tutorials.nemo-retriever-synthetic-data-generation.config.config"
)
RetrieverEvalSDGConfig = config.RetrieverEvalSDGConfig


def get_pipeline(args: Any) -> Any:

    cfg = RetrieverEvalSDGConfig.from_yaml(args.pipeline_config)
    # update api_key from input args
    cfg.api_key = args.api_key

    if args.pipeline_type == "generate":
        sdg_pipeline = Sequential(
            [
                RetrieverEvalSetGenerator(cfg),
            ]
        )
    else:
        sdg_pipeline = None

    filters = []

    if args.pipeline_type == "filter":
        if cfg.easiness_filter:
            filters.append(
                ScoreFilter(
                    EasinessFilter(
                        cfg.base_url,
                        cfg.api_key,
                        cfg.easiness_filter,
                        cfg.percentile,
                        cfg.truncate,
                        cfg.batch_size,
                    ),
                    text_field=["text", "question"],
                    score_field="easiness_scores",
                )
            )
        if cfg.answerability_filter:
            filters.append(
                ScoreFilter(
                    AnswerabilityFilter(
                        cfg.base_url,
                        cfg.api_key,
                        cfg.answerability_filter,
                        cfg.answerability_system_prompt,
                        cfg.answerability_user_prompt_template,
                        cfg.num_criteria,
                    ),
                    text_field=["text", "question"],
                    score_field="answerability_scores",
                )
            )
    if filters:
        filtering_pipeline = Sequential(filters)
    else:
        filtering_pipeline = None

    return sdg_pipeline, filtering_pipeline


def write_to_beir(args: Any, dataset: DocumentDataset, input_dataset: DocumentDataset):

    df = dataset.df
    df = df.compute()

    save_dir = os.path.join(args.output_dir, "beir")
    qrels_save_dir = os.path.join(args.output_dir, "beir", "qrels")

    os.makedirs(save_dir)
    os.makedirs(qrels_save_dir)

    corpus_save_path = os.path.join(args.output_dir, "beir", "corpus.jsonl")
    queries_save_path = os.path.join(args.output_dir, "beir", "queries.jsonl")
    df[["question-id", "question"]].rename(
        columns={"question-id": "_id", "question": "text"}
    ).to_json(queries_save_path, lines=True, orient="records")

    df[["question-id", "_id", "score"]].rename(
        columns={"question-id": "query-id", "_id": "corpus-id"}
    ).to_csv(os.path.join(qrels_save_dir, "test.tsv"), sep="\t", index=False)

    if args.pipeline_type == "filter":
        input_df = input_dataset.df.compute()
        input_df = input_df.groupby("_id").agg({"text": set}).reset_index()
        input_df["text"] = input_df["text"].map(lambda x: x.pop())
        input_df[["_id", "text"]].to_json(
            corpus_save_path, lines=True, orient="records"
        )
    elif args.pipeline_type == "generate":
        df = df.groupby("_id").agg({"text": set}).reset_index()
        df["text"] = df["text"].map(lambda x: x.pop())
        df[["_id", "text"]].to_json(corpus_save_path, lines=True, orient="records")


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "--input-dir",
        type=str,
        default="",
        help="Input directory containing jsonl files that have the document/text chunks for query & answer generation",
    )
    parser.add_argument(
        "--input-format",
        type=str,
        default="jsonl",
        help="The input files must be in jsonl format",
    )
    parser.add_argument(
        "--pipeline-config",
        type=str,
        default="",
        help="Pipeline configuartion yaml file path",
    )
    parser.add_argument(
        "--output-dir",
        type=str,
        default="",
        help="Output dir for generated data",
    )
    parser.add_argument(
        "--api-key",
        type=str,
        default=None,
        help="The API key to use for the synthetic data generation LLM client.",
    )
    parser.add_argument(
        "--api-timeout",
        type=int,
        default=120,
        help="The timeout value for API calls in seconds.",
    )
    parser.add_argument(
        "--n-partitions",
        type=int,
        default=1,
        help="Number of partitions for parallel processing of data.",
    )
    parser.add_argument(
        "--pipeline-type",
        type=str,
        default="generate",
        help="Choices: 1. 'generate', 2. 'filter'",
    )
    parser.add_argument(
        "--save-format",
        type=str,
        default="jsonl",
        help="Save format choices 1. beir 2. jsonl",
    )

    args = parser.parse_args()

    if not os.path.exists(args.output_dir):
        os.makedirs(args.output_dir)
    elif not any(os.scandir(args.output_dir)):
        print("Provided directory exists but is empty, using the empty directory")
    else:
        raise ValueError("Output directory exists already, use a new directory!")

    if args.input_format == "jsonl":
        if args.pipeline_type == "filter":
            input_files = get_all_files_paths_under(
                args.input_dir, keep_extensions="part"
            )
        elif args.pipeline_type == "generate":
            input_files = get_all_files_paths_under(
                args.input_dir, keep_extensions="jsonl"
            )
        else:
            raise ValueError(
                "Error only two pipelines supported: 'generate' & 'filter'"
            )

        input_dataset = DocumentDataset.read_json(input_files)
    else:
        raise ValueError("Error: Only jsonl format supported")

    if args.n_partitions:
        ddf = input_dataset.df
        n_data = len(ddf)
        if args.n_partitions < n_data:
            ddf = ddf.repartition(npartitions=args.n_partitions)
            input_dataset = DocumentDataset(ddf)
        else:
            print("Number of partitions greater than data size, using 1 partition")

    sdg_pipeline, filtering_pipeline = get_pipeline(args)

    if sdg_pipeline:
        print("Generating data ...")
        st_time = time.time()
        generated_dataset = sdg_pipeline(input_dataset)
        generated_dataset.persist()

        if args.save_format == "jsonl":
            print("Writing all generated data to disk ...")
            all_save_dir = os.path.join(args.output_dir, "jsonl")
            os.makedirs(all_save_dir)
            generated_dataset.to_json(all_save_dir)

        # saving in beir format
        if args.save_format == "beir":
            print("Write all data in beir format")
            write_to_beir(args, generated_dataset, input_dataset)

        print("Time taken to generate data = {:.2f} s".format(time.time() - st_time))

    if filtering_pipeline:
        print("Filtering data ...")
        st_time = time.time()
        filtered_dataset = filtering_pipeline(input_dataset)
        filtered_dataset.persist()

        if args.save_format == "jsonl":
            print("Writing filtered data to disk ...")
            all_save_dir = os.path.join(args.output_dir, "jsonl")
            os.makedirs(all_save_dir)
            filtered_dataset.to_json(all_save_dir)

        if args.save_format == "beir":
            print("Writing filtered data in beir format")
            # saving in beir format
            write_to_beir(args, filtered_dataset, input_dataset)

        print("Time taken to filter data = {:.2f} s".format(time.time() - st_time))

    print("RUN complete!")
    print("------------------------")


if __name__ == "__main__":
    dask_client = get_client()
    main()



# Copyright (c) 2025, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import pandas as pd
import pytest
from dask.dataframe.utils import assert_eq

from nemo_curator import (
    BaseModule,
    FuzzyDuplicates,
    FuzzyDuplicatesConfig,
    ScoreFilter,
    Sequential,
    ToBackend,
)
from nemo_curator.datasets import DocumentDataset
from nemo_curator.filters import MeanWordLengthFilter
from nemo_curator.utils.import_utils import gpu_only_import

cudf = gpu_only_import("cudf")
dask_cudf = gpu_only_import("dask_cudf")


class CPUModule(BaseModule):
    def __init__(self):
        super().__init__(input_backend="pandas")

    def call(self, dataset: DocumentDataset):
        dataset.df["cpu_lengths"] = dataset.df["text"].str.len()
        return dataset


class GPUModule(BaseModule):
    def __init__(self):
        super().__init__(input_backend="cudf")

    def call(self, dataset: DocumentDataset):
        dataset.df["gpu_lengths"] = dataset.df["text"].str.len()
        return dataset


class AnyModule(BaseModule):
    def __init__(self):
        super().__init__(input_backend="any")

    def call(self, dataset: DocumentDataset):
        dataset.df["any_lengths"] = dataset.df["text"].str.len()
        return dataset


@pytest.fixture
def raw_data():
    base_data = {
        "id": [1, 2, 3, 4, 100, 200, 300],
        "text": [
            "The quick brown fox jumps over the lazy dog",
            "The quick brown foxes jumps over the lazy dog",
            "The quick brown wolf jumps over the lazy dog",
            "The quick black cat jumps over the lazy dog",
            "A test string",
            "Another test string",
            "A different object",
        ],
    }
    gt_results = [43, 45, 44, 43, 13, 19, 18]

    return base_data, gt_results


@pytest.fixture
def cpu_data(raw_data):
    base_data, gt_results = raw_data
    df = pd.DataFrame(base_data)
    gt_lengths = pd.Series(gt_results, name="cpu_lengths")
    return DocumentDataset.from_pandas(df), gt_lengths


@pytest.fixture
def gpu_data(raw_data):
    base_data, gt_results = raw_data
    df = cudf.DataFrame(base_data)
    df = dask_cudf.from_cudf(df, 2)
    gt_lengths = cudf.Series(gt_results, name="gpu_lengths", dtype="int32")
    return DocumentDataset(df), gt_lengths


@pytest.mark.gpu
class TestBackendSupport:
    def test_pandas_backend(
        self,
        cpu_data,
        gpu_client,
    ):
        print("client", gpu_client)
        dataset, gt_lengths = cpu_data
        pipeline = CPUModule()
        result = pipeline(dataset)
        result_df = result.df.compute()
        assert_eq(result_df["cpu_lengths"], gt_lengths)

    def test_cudf_backend(
        self,
        gpu_data,
        gpu_client,
    ):
        print("client", gpu_client)
        dataset, gt_lengths = gpu_data
        pipeline = GPUModule()
        result = pipeline(dataset)
        result_df = result.df.compute()
        assert_eq(result_df["gpu_lengths"], gt_lengths)

    def test_any_backend(
        self,
        cpu_data,
        gpu_data,
        gpu_client,
    ):
        print("client", gpu_client)
        cpu_dataset, gt_cpu_lengths = cpu_data
        gt_cpu_lengths = gt_cpu_lengths.rename("any_lengths")
        gpu_dataset, gt_gpu_lengths = gpu_data
        gt_gpu_lengths = gt_gpu_lengths.rename("any_lengths")
        pipeline = AnyModule()

        cpu_result = pipeline(cpu_dataset)
        cpu_result_df = cpu_result.df.compute()
        assert_eq(cpu_result_df["any_lengths"], gt_cpu_lengths)
        gpu_result = pipeline(gpu_dataset)
        gpu_result_df = gpu_result.df.compute()
        assert_eq(gpu_result_df["any_lengths"], gt_gpu_lengths)

    def test_pandas_to_cudf(
        self,
        cpu_data,
        gpu_data,
        gpu_client,
    ):
        print("client", gpu_client)
        dataset, gt_cpu_lengths = cpu_data
        _, gt_gpu_lengths = gpu_data
        pipeline = Sequential(
            [
                CPUModule(),
                ToBackend("cudf"),
                GPUModule(),
            ]
        )
        result = pipeline(dataset)
        result_df = result.df.compute()
        assert_eq(result_df["cpu_lengths"], gt_cpu_lengths)
        assert_eq(result_df["gpu_lengths"], gt_gpu_lengths)

    def test_cudf_to_pandas(
        self,
        cpu_data,
        gpu_data,
        gpu_client,
    ):
        print("client", gpu_client)
        _, gt_cpu_lengths = cpu_data
        dataset, gt_gpu_lengths = gpu_data
        pipeline = Sequential(
            [
                GPUModule(),
                ToBackend("pandas"),
                CPUModule(),
            ]
        )
        result = pipeline(dataset)
        result_df = result.df.compute()
        assert_eq(result_df["cpu_lengths"], gt_cpu_lengths)
        assert_eq(result_df["gpu_lengths"], gt_gpu_lengths)

    def test_5x_switch(
        self,
        cpu_data,
        gpu_data,
        gpu_client,
    ):
        print("client", gpu_client)
        dataset, gt_cpu_lengths = cpu_data
        _, gt_gpu_lengths = gpu_data
        pipeline = Sequential(
            [
                CPUModule(),
                ToBackend("cudf"),
                GPUModule(),
                ToBackend("pandas"),
                CPUModule(),
                ToBackend("cudf"),
                GPUModule(),
                ToBackend("pandas"),
                CPUModule(),
                ToBackend("cudf"),
                GPUModule(),
            ]
        )
        result = pipeline(dataset)
        result_df = result.df.compute()
        assert sorted(list(result_df.columns)) == [
            "cpu_lengths",
            "gpu_lengths",
            "id",
            "text",
        ]
        assert_eq(result_df["cpu_lengths"], gt_cpu_lengths)
        assert_eq(result_df["gpu_lengths"], gt_gpu_lengths)

    def test_wrong_backend_cpu_data(self, cpu_data, gpu_client):
        with pytest.raises(ValueError):
            print("client", gpu_client)
            dataset, _ = cpu_data
            pipeline = GPUModule()
            result = pipeline(dataset)
            _ = result.df.compute()

    def test_wrong_backend_gpu_data(self, gpu_data, gpu_client):
        with pytest.raises(ValueError):
            print("client", gpu_client)
            dataset, _ = gpu_data
            pipeline = CPUModule()
            result = pipeline(dataset)
            _ = result.df.compute()

    def test_unsupported_to_backend(self, cpu_data, gpu_client):
        with pytest.raises(ValueError):
            print("client", gpu_client)
            dataset, _ = cpu_data
            pipeline = ToBackend("fake_backend")
            result = pipeline(dataset)
            _ = result.df.compute()


@pytest.fixture
def real_module_raw_data():
    base_data = {
        "id": [1, 2, 3, 4, 100, 200, 300],
        "text": [
            "The quick brown fox jumps over the lazy dog",
            "The quick brown foxes jumps over the lazy dog",
            "The quick brown wolf jumps over the lazy dog",
            "The quick black cat jumps over the lazy dog",
            "A test string",
            "Another test string",
            "A different object",
        ],
    }
    return base_data


@pytest.fixture
def real_module_cpu_data(real_module_raw_data):
    df = pd.DataFrame(real_module_raw_data)
    gt_results = pd.Series(
        [35 / 9, 37 / 9, 4.0, 35 / 9, 33 / 9, 51 / 9, 48 / 9], name="mean_lengths"
    )
    return DocumentDataset.from_pandas(df), gt_results


@pytest.fixture
def real_module_gpu_data(real_module_raw_data):
    df = cudf.DataFrame(real_module_raw_data)
    df = dask_cudf.from_cudf(df, 2)
    gt_results = cudf.Series([[1, 2, 3, 4], [100, 200]], name="id")
    return DocumentDataset(df), gt_results


@pytest.mark.gpu
class TestRealModules:
    def test_score_filter(
        self,
        real_module_cpu_data,
        gpu_client,
    ):
        print("client", gpu_client)
        dataset, gt_results = real_module_cpu_data
        pipeline = ScoreFilter(
            MeanWordLengthFilter(), score_field="mean_lengths", score_type=float
        )
        result = pipeline(dataset)
        result_df = result.df.compute()
        assert_eq(result_df["mean_lengths"], gt_results)

    def test_score_filter_wrong_backend(
        self,
        real_module_gpu_data,
        gpu_client,
    ):
        with pytest.raises(ValueError):
            print("client", gpu_client)
            dataset, _ = real_module_gpu_data
            pipeline = ScoreFilter(
                MeanWordLengthFilter(), score_field="mean_lengths", score_type=float
            )
            result = pipeline(dataset)
            _ = result.df.compute()

    def test_fuzzy_dedup(
        self,
        real_module_gpu_data,
        tmpdir,
        gpu_client,
    ):
        print(gpu_client)
        dataset, gt_results = real_module_gpu_data
        # Dedup might fail when indices per partition do not start from 0
        dataset.df = dataset.df.reset_index(drop=True)
        config = FuzzyDuplicatesConfig(
            cache_dir=tmpdir,
            id_field="id",
            text_field="text",
            seed=42,
            char_ngrams=5,
            num_buckets=15,
            hashes_per_bucket=1,
            use_64_bit_hash=False,
            buckets_per_shuffle=3,
            false_positive_check=True,
            num_anchors=2,
            jaccard_threshold=0.3,
        )
        fuzzy_duplicates = FuzzyDuplicates(config=config)
        result = fuzzy_duplicates(dataset)
        result_df = result.df.compute()
        # Drop non duplicated docs
        result_df = result_df[result_df.group.duplicated(keep=False)]
        result_df = result_df.groupby("group").id.agg(list)
        # Sort to maintain uniform ordering

        result_df = result_df.list.sort_values()
        result_df = result_df.sort_values()
        gt_results = gt_results.list.sort_values()
        gt_results = gt_results.sort_values()
        assert_eq(gt_results, result_df, check_index=False)

    def test_fuzzy_dedup_wrong_backend(
        self,
        real_module_cpu_data,
        tmpdir,
        gpu_client,
    ):
        with pytest.raises(ValueError):
            print(gpu_client)
            dataset, _ = real_module_cpu_data
            # Dedup might fail when indices per partition do not start from 0
            dataset.df = dataset.df.reset_index(drop=True)
            config = FuzzyDuplicatesConfig(
                cache_dir=tmpdir,
                id_field="id",
                text_field="text",
                seed=42,
                char_ngrams=5,
                num_buckets=15,
                hashes_per_bucket=1,
                use_64_bit_hash=False,
                buckets_per_shuffle=3,
                false_positive_check=True,
                num_anchors=2,
                jaccard_threshold=0.3,
            )
            fuzzy_duplicates = FuzzyDuplicates(config=config)
            result = fuzzy_duplicates(dataset)
            _ = result.df.compute()

    def test_score_filter_and_fuzzy(
        self,
        real_module_cpu_data,
        real_module_gpu_data,
        tmpdir,
        gpu_client,
    ):
        print("client", gpu_client)
        dataset, _ = real_module_cpu_data
        _, gt_results = real_module_gpu_data
        dataset.df = dataset.df.reset_index(drop=True)
        config = FuzzyDuplicatesConfig(
            cache_dir=tmpdir,
            id_field="id",
            text_field="text",
            seed=42,
            char_ngrams=5,
            num_buckets=15,
            hashes_per_bucket=1,
            use_64_bit_hash=False,
            buckets_per_shuffle=3,
            false_positive_check=True,
            num_anchors=2,
            jaccard_threshold=0.3,
        )
        pipeline = Sequential(
            [
                ScoreFilter(
                    MeanWordLengthFilter(), score_field="mean_lengths", score_type=float
                ),
                ToBackend("cudf"),
                FuzzyDuplicates(config=config),
            ]
        )

        result = pipeline(dataset)
        result_df = result.df.compute()
        # Right now the output of FuzzyDuplicates does not retain the original metadata
        # so we simply check the output of fuzzy dedupe to ensure accuracy
        # Drop non duplicated docs
        result_df = result_df[result_df.group.duplicated(keep=False)]
        result_df = result_df.groupby("group").id.agg(list)
        # Sort to maintain uniform ordering
        result_df = result_df.list.sort_values()
        result_df = result_df.sort_values()
        gt_results = gt_results.list.sort_values()
        gt_results = gt_results.sort_values()
        assert_eq(gt_results, result_df, check_index=False)



# Copyright (c) 2024, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from nemo_curator.filters import DocumentFilter


class IncompleteStoryFilter(DocumentFilter):
    """
    If the document doesn't end with a terminating punctuation mark, then discard.
    """

    def __init__(self):
        super().__init__()
        # Accepted story terminators.
        self._story_terminators = {".", "!", "?", '"', "”"}

    def score_document(self, text: str) -> bool:
        """
        Determines if a document's score is valid based on the last character of the text.

        Args:
            text (str): The document text.

        Returns:
            bool: True if the document's score is valid, False otherwise.
        """
        ret = text.strip()[-1] in self._story_terminators
        return ret

    def keep_document(self, score) -> bool:
        return score



# Copyright (c) 2024, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from abc import ABC, abstractmethod
from typing import Dict, List, Optional, Union

from dask.typing import no_default

from nemo_curator.datasets.parallel_dataset import ParallelDataset
from nemo_curator.utils.module_utils import is_batched


class BitextFilter(ABC):
    """A base class for bitext filter objects (such as length ratio, QE filter) on bitext.
    Different from `ParallelScoreFilter`, these filters require looking at both source AND target side of the bitext to compute a score.

    This is roughly equivalent to a `ScoreFilter` wrapping over a `DocumentFilter` object.
    But aside from operating on `ParallelDataset` instead of `DocumentDataset`, it comes with some other differences:

    - It discarded the ScoreFilter/DocumentFilter hierarchy. So filter classes can directly be used instead of being wrapped by ScoreFilter.
    - Unlike an DocumentFilter object, it allows passing extra metadata information into the scoring function.
    """

    def __init__(
        self,
        src_field: str = "src",
        tgt_field: str = "tgt",
        metadata_fields: Union[List[str], str] = [],
        metadata_field_name_mapping: Dict[str, str] = {},
        score_field: Optional[str] = None,
        score_type: Union[type, str] = None,
        invert=False,
    ):
        """Args:
            src_field (str, optional): The field the source documents will be read from. Defaults to "src".
            tgt_field (str, optional): The field the target documents will be read from. Defaults to "tgt".
            metadata_fields (Union[List[str], str], optional): Name of the metadata fields in case fields other than source and target documents need to be accessed. Defaults to [].
            metadata_field_name_mapping (Dict[str, str], optional): Mapping of field names in the data to argument names in `_score_bitext` function, in case they are different.
                For example, if a field is called "src" in the data but should be passed to an argument called "source" in `_score_bitext` function,
                you should add an entry `{"src": "source"}`. Identity map is assumed if a mapping is not specified for a field name. Default to {}.
            score_field (Optional[str], optional): The field to which the scores will be written. If None, scores will be immediately discarded after use. Defaults to None.
            score_type (Union[type, str], optional): The datatype of the score that will be made for each document. Defaults to None.
            invert (bool, optional): If True, will keep all documents that are normally discarded. Defaults to False.

        Raises:
            ValueError: If length of source and target fields are different.
        """

        self.src_field = src_field
        self.tgt_field = tgt_field
        self.metadata_fields = metadata_fields
        self.metadata_field_name_mapping = metadata_field_name_mapping
        self.score_field = score_field
        self.score_type = score_type
        self.invert = invert

    def __call__(
        self,
        dataset: ParallelDataset,
    ) -> ParallelDataset:
        """Scores and filters all records in the dataset

        Args:
            dataset (ParallelDataset): The dataset to apply the module to.

        Returns:
            ParallelDataset:  A dataset with the score and filter applied
        """
        # Set the metadata for the function calls if provided
        if self.score_type:
            meta = (None, self.score_type)
        else:
            meta = no_default

        # support multiple fields if supplied
        fields = []
        fields.append(self.src_field)
        fields.append(self.tgt_field)
        fields.extend(self.metadata_fields)

        if is_batched(self.score_bitext):
            scores = dataset.df[fields].map_partitions(
                self._score_bitext_wrapper,
                metadata_field_name_mapping=self.metadata_field_name_mapping,
                meta=meta,
            )
        else:
            scores = dataset.df[fields].apply(
                self._score_bitext_wrapper,
                metadata_field_name_mapping=self.metadata_field_name_mapping,
                axis=1,
                meta=meta,
            )

        if self.score_field is not None:
            dataset.df[self.score_field] = scores

        if is_batched(self.keep_bitext):
            bool_mask = scores.map_partitions(self.keep_bitext, meta=(None, bool))
        else:
            bool_mask = scores.apply(self.keep_bitext, meta=(None, bool))
        if self.invert:
            bool_mask = ~bool_mask

        return ParallelDataset(dataset.df[bool_mask])

    def _score_bitext_wrapper(
        self,
        df,
        metadata_field_name_mapping: Dict[str, str] = {},
    ):
        """In the batch mode, pass fields in a data frame to arguments of a function, according to a name mapping.

        Args:
            df (DataFrame): data frame to perform the mapping on.
            metadata_field_name_mapping (Dict[str, str]): see `__call__` function for details.
        """
        kwargs = {}
        kwargs["src"] = df[self.src_field]
        kwargs["tgt"] = df[self.tgt_field]
        for field_name in self.metadata_fields:
            arg_name = metadata_field_name_mapping.get(field_name, field_name)
            kwargs[arg_name] = df[field_name]

        return self.score_bitext(**kwargs)

    @abstractmethod
    def score_bitext(self, src, tgt, **kwargs):
        """Scoring function for the bitext."""
        pass

    @abstractmethod
    def keep_bitext(self, **kwargs):
        pass



# Copyright (c) 2024, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from __future__ import annotations

import warnings
from typing import Iterable, List, Optional, Union

from nemo_curator.services.conversation_formatter import ConversationFormatter

from .model_client import AsyncLLMClient, LLMClient


class NemoDeployClient(LLMClient):
    """
    A wrapper around NemoQueryLLM for querying models in synthetic data generation
    """

    def __init__(self, nemo_deploy: NemoQueryLLM) -> None:
        self.client = nemo_deploy

    def query_model(
        self,
        *,
        messages: Iterable,
        model: str,
        conversation_formatter: Optional[ConversationFormatter] = None,
        max_tokens: Optional[int] = None,
        n: Optional[int] = None,
        seed: Optional[int] = None,
        stop: Union[Optional[str], List[str]] = [],
        stream: bool = False,
        temperature: Optional[float] = None,
        top_k: Optional[int] = None,
        top_p: Optional[float] = None,
    ) -> List[str]:
        if conversation_formatter is None:
            raise ValueError(
                "NemoDeployClient's query_model requires a conversation_formatter"
            )

        prompt = conversation_formatter.format_conversation(messages)
        self.client.model_name = model

        if n is not None:
            warnings.warn("n is not supported in NemoDeployClient")
        if stream:
            warnings.warn("streamming is not supported in NeMoDeployClient")

        if isinstance(stop, str):
            stop = [stop]

        response = self.client.query_llm(
            prompts=[prompt],
            max_output_len=max_tokens,
            random_seed=seed,
            stop_words_list=stop,
            temperature=temperature,
            top_p=top_p,
            top_k=top_k,
        )[0]

        return self._postprocess_response(response, stop)

    @staticmethod
    def _postprocess_response(responses: List[str], stop_words: List[str]) -> List[str]:
        processed_responses = []
        for response in responses:
            for stop in stop_words:
                if response.endswith(stop):
                    response = response[: -len(stop)]
            processed_responses.append(response.strip())
        return processed_responses

    def query_reward_model(self, *, messages: Iterable, model: str) -> dict:
        """
        Prompts an LLM Reward model to score a conversation between a user and assistant
        Args:
            messages: The conversation to calculate a score for.
                Should be formatted like:
                    [{"role": "user", "content": "Write a sentence"}, {"role": "assistant", "content": "This is a sentence"}, ...]
            model: The name of the model that should be used to calculate the reward.
                Must be a reward model, cannot be a regular LLM.
        Returns:
            A mapping of score_name -> score
        """
        raise NotImplementedError(
            "Reward model inference is not supported in NeMo Deploy Clients"
        )



import logging
import os
import time

import cudf
import dask_cudf
import numpy as np

from nemo_curator import LSH
from nemo_curator.datasets import DocumentDataset
from nemo_curator.utils.distributed_utils import get_client, get_num_workers
from nemo_curator.utils.fuzzy_dedup_utils.id_mapping import convert_str_id_to_int

logging.basicConfig(format="%(asctime)s: %(message)s", level=logging.INFO)


DATA_BASE = os.environ.get("DATA_BASE")
SCHEDULER_FILE = os.environ.get("SCHEDULER_FILE")


if __name__ == "__main__":
    client = get_client(scheduler_file=SCHEDULER_FILE)
    logging.info(f"Number of dask workers: {get_num_workers(client)}")

    minhash_base_output_path = os.path.join(DATA_BASE, "fuzzy/minhash")
    minhash_output_dir = os.path.join(minhash_base_output_path, "data")

    # Input
    lsh_input_data_path = minhash_output_dir

    # Output
    lsh_base_output_path = os.path.join(DATA_BASE, "fuzzy/lsh")
    lsh_output_dir = os.path.join(lsh_base_output_path, "data")

    # Relevant parameters
    lsh_id_field = "nemo_id"
    minhash_field = "_minhash_signature"
    minhash_length = 128
    num_bands = 8
    buckets_per_shuffle = 8

    t0 = time.time()

    # Load MinHash output
    logging.info("Converting ids")
    df = dask_cudf.read_parquet(lsh_input_data_path, backend="cudf")
    df = df.map_partitions(
        convert_str_id_to_int,
        id_column=lsh_id_field,
        meta=cudf.DataFrame(
            {minhash_field: [[1, 2, 3]], "doc_id": [1], "dataset_id": np.uint32(1)}
        ),
    )
    # Run LSH()
    lsh = LSH(
        cache_dir=lsh_output_dir,
        num_hashes=minhash_length,
        num_buckets=num_bands,
        buckets_per_shuffle=buckets_per_shuffle,
        id_fields=["dataset_id", "doc_id"],
        minhash_field=minhash_field,
    )
    res = lsh(DocumentDataset(df))

    t1 = time.time()
    logging.info(f"Time taken for LSH: {time.time() - t0} s")



import json
import logging
import os

import cudf
import dask_cudf
import tqdm

from nemo_curator.utils.distributed_utils import get_client, get_num_workers

logging.basicConfig(format="%(asctime)s: %(message)s", level=logging.INFO)


def list_deepest_folders(path):
    deepest_folders = []

    # Walk through the directory structure
    for root, dirs, files in os.walk(
        path, topdown=False
    ):  # `topdown=False` starts from the deepest folder
        if not dirs:  # Keep only folders that don't have subdirectories
            deepest_folders.append(root)

    return deepest_folders


def read_data_subset(dir_paths: list):
    """
    Reads 1 row from each dataset assuming each file has a unique dataset prefix
    """
    dfs = []
    for dir_path in tqdm.tqdm(dir_paths):
        file_name = sorted([x for x in os.listdir(dir_path) if ".parquet" in x])[0]
        file_path = os.path.join(dir_path, file_name)
        x = dask_cudf.read_parquet(file_path).head(1)  # read 1 rows from each file
        dfs.append(x)
    x = cudf.concat(dfs)
    return x


def generate_mapping(x: cudf.DataFrame):
    dx = x.nemo_id.str.rsplit("-", n=1, expand=True)
    x["dataset"] = dx[0]
    x["dataset_id"] = x.dataset.hash_values()
    mapping_df = x[["dataset", "dataset_id"]]
    mapping_df = mapping_df.drop_duplicates()
    mapping_df["dataset_id"] = mapping_df.dataset_id.astype(str)
    dataset_id_mapping = mapping_df.set_index("dataset_id")["dataset"].to_dict()
    return dataset_id_mapping


def convert_cc_ids(cc_df: dask_cudf.DataFrame, doc_id_mapping: dict, pad_width=10):
    cc_df["doc_id"] = (
        cc_df["doc_id"].astype(str).str.pad(width=pad_width, side="left", fillchar="0")
    )
    cc_df["dataset_id"] = cc_df.dataset_id.astype(str).replace(doc_id_mapping)
    cc_df["original_id"] = cc_df.dataset_id + "-" + cc_df.doc_id
    return cc_df[["original_id", "group"]]


DATA_BASE = os.environ.get("DATA_BASE")
SCHEDULER_FILE = os.environ.get("SCHEDULER_FILE")
ID_MAPPING = os.path.join(DATA_BASE, "dataset_id_mapping.json")


if __name__ == "__main__":
    client = get_client(scheduler_file=SCHEDULER_FILE)
    logging.info(f"Number of dask workers: {get_num_workers(client)}")

    all_folders = sorted(list_deepest_folders(os.path.join(DATA_BASE, "processed")))
    df_subset = read_data_subset(all_folders)
    dataset_id_mapping = generate_mapping(df_subset)

    with open(ID_MAPPING, "w") as f:
        f.write(json.dumps(dataset_id_mapping))



# Copyright (c) 2024, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import argparse
import random

import fasttext

import nemo_curator as nc
from nemo_curator.datasets import DocumentDataset
from nemo_curator.filters import FastTextQualityFilter
from nemo_curator.modifiers import FastTextLabelModifier
from nemo_curator.utils.distributed_utils import get_client, read_data, write_to_disk
from nemo_curator.utils.file_utils import get_all_files_paths_under
from nemo_curator.utils.script_utils import ArgumentHelper


def load_dataset(input_data_dir):
    files = list(get_all_files_paths_under(input_data_dir, keep_extensions="jsonl"))
    raw_data = read_data(files, file_type="jsonl", backend="pandas", add_filename=True)
    dataset = DocumentDataset(raw_data)

    return dataset


def create_samples(data_path, label, num_samples):
    raw_dataset = load_dataset(data_path)
    label_quality = nc.Modify(FastTextLabelModifier(label))

    labeled_dataset = label_quality(raw_dataset)
    labeled_samples = labeled_dataset.df.sample(
        frac=num_samples / len(labeled_dataset.df)
    )

    return labeled_samples["text"].compute().values.tolist()


def main(args):
    # Params
    low_quality_data_path = "/path/to/low_quality"
    high_quality_data_path = "/path/to/high_quality"
    num_low_quality_samples = 1000
    num_high_quality_samples = 1000
    filtered_output = "/path/to/output"

    # Prepare samples for the classifier
    client = get_client(**ArgumentHelper.parse_client_args(args))
    low_quality_samples = create_samples(
        low_quality_data_path, "__label__lq", num_low_quality_samples
    )
    high_quality_samples = create_samples(
        high_quality_data_path, "__label__hq", num_high_quality_samples
    )

    train_samples = low_quality_samples + high_quality_samples
    random.shuffle(train_samples)
    train_file = "./fasttext.train"
    model_path = "./fasttext_model.bin"
    with open(train_file, "w") as f:
        for sample in train_samples:
            f.write(sample)
            f.write("\n")

    # Train fastText classifier
    model = fasttext.train_supervised(
        input=train_file,
        lr=0.01,
        dim=100,
        epoch=5,
        wordNgrams=2,
    )
    model.save_model(model_path)

    # Filter data
    target_dataset = load_dataset(low_quality_data_path)
    filter_pipeline = nc.ScoreFilter(
        FastTextQualityFilter(model_path),
        score_field="quality_score",
        score_type=float,
    )
    filtered_dataset = filter_pipeline(target_dataset)

    # Write filtered dataset
    write_to_disk(filtered_dataset.df, filtered_output, write_to_filename=True)


def attach_args(
    parser=argparse.ArgumentParser(
        formatter_class=argparse.ArgumentDefaultsHelpFormatter
    ),
):
    return ArgumentHelper(parser).add_distributed_args()


if __name__ == "__main__":
    main(attach_args().parse_args())



# Copyright (c) 2024, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import os
from dataclasses import dataclass

os.environ["RAPIDS_NO_INITIALIZE"] = "1"
from abc import abstractmethod
from typing import List, Optional, Union

import torch
import torch.nn as nn
from crossfit import op
from huggingface_hub import PyTorchModelHubMixin
from transformers import AutoModel

from nemo_curator.datasets import DocumentDataset
from nemo_curator.modules.base import BaseModule
from nemo_curator.utils.distributed_utils import get_gpu_memory_info


class DistributedDataClassifier(BaseModule):
    """Abstract class for running multi-node multi-GPU data classification"""

    def __init__(
        self,
        model: str,
        labels: Optional[List[str]],
        filter_by: Optional[List[str]],
        batch_size: int,
        out_dim: Optional[int],
        pred_column: Union[str, List[str]],
        max_chars: int,
        device_type: str,
        autocast: bool,
    ):
        super().__init__(input_backend="cudf")
        self.model = model
        self.labels = labels
        self.filter_by = filter_by
        self.batch_size = batch_size
        self.out_dim = out_dim
        self.pred_column = pred_column
        self.max_chars = max_chars
        self.device_type = device_type
        self.autocast = autocast

    def call(self, dataset: DocumentDataset) -> DocumentDataset:
        result_doc_dataset = self._run_classifier(dataset)
        if self.filter_by is not None:
            return self._filter_documents(result_doc_dataset)

        return result_doc_dataset

    @abstractmethod
    def _run_classifier(self) -> DocumentDataset:
        pass

    def _filter_documents(
        self,
        dataset: DocumentDataset,
    ) -> DocumentDataset:
        df = dataset.df

        filter_by = self.filter_by
        if type(filter_by) == str:
            filtered_df = df[df[self.pred_column].astype(str) == filter_by]
            return DocumentDataset(filtered_df)
        elif type(filter_by) == list:
            filtered_df = df[df[self.pred_column].isin(filter_by)]
            return DocumentDataset(filtered_df)

        raise TypeError("filter_by must be a string or list type")

    def get_labels(self) -> List[str]:
        return self.labels


class HFDeberta(nn.Module, PyTorchModelHubMixin):
    def __init__(self, config: dataclass):
        super(HFDeberta, self).__init__()
        self.model = AutoModel.from_pretrained(config["base_model"])
        self.dropout = nn.Dropout(config["fc_dropout"])
        self.fc = nn.Linear(self.model.config.hidden_size, len(config["id2label"]))

    def _forward(self, batch):
        features = self.model(
            batch["input_ids"], batch["attention_mask"]
        ).last_hidden_state
        dropped = self.dropout(features)
        outputs = self.fc(dropped)
        return torch.softmax(outputs[:, 0, :], dim=1)

    def forward(self, batch):
        if self.autocast:
            with torch.autocast(device_type="cuda"):
                return self._forward(batch)
        else:
            return self._forward(batch)

    def set_autocast(self, autocast: bool):
        self.autocast = autocast


def _run_classifier_helper(
    df: "dask_cudf.DataFrame",
    model: "HFModel",
    labels: list[str],
    max_chars: int,
    batch_size: int,
    label_col: str,
    text_field: str = "text",
    prob_col: str = None,
) -> "dask_cudf.DataFrame":

    if prob_col is None:
        prob_col = "_prob"
        labeler = op.Labeler(labels, cols=[prob_col], suffix=label_col)
    else:
        labeler = op.Labeler(
            labels, cols=[prob_col], keep_cols=[prob_col], suffix=label_col
        )

    columns_to_keep_list = df.columns.to_list()

    classifier_pipe = op.Sequential(
        op.Tokenizer(
            model, cols=[text_field], tokenizer_type="default", max_chars=max_chars
        ),
        op.Predictor(
            model,
            sorted_data_loader=True,
            batch_size=batch_size,
            pred_output_col=prob_col,
        ),
        labeler,
        repartition=df.npartitions,
        keep_cols=columns_to_keep_list,
    )
    df = classifier_pipe(df)

    return df


def _get_suggest_memory_for_classifier() -> int:
    gpu_memory_info = get_gpu_memory_info()
    min_gpu_memory = min(gpu_memory_info.values())
    # Convert memory from bytes to GB
    min_gpu_memory_gb = min_gpu_memory / (1024**3)
    # Subtract 4GB from the minimum
    # to leave room for other operations
    # like cuDF operations
    min_gpu_memory_gb = min_gpu_memory_gb - 4
    return int(min_gpu_memory_gb)



# Copyright (c) 2025, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import importlib
import itertools
from typing import TYPE_CHECKING

import numpy as np
import pandas as pd
from dask.base import normalize_token
from openai import OpenAI

from nemo_curator import ClusteringModel
from nemo_curator.datasets import DocumentDataset
from nemo_curator.utils.distributed_utils import load_object_on_worker

config = importlib.import_module(
    "tutorials.nemo-retriever-synthetic-data-generation.config.config"
)
RetrieverHardNegativeMiningConfig = config.RetrieverHardNegativeMiningConfig

if TYPE_CHECKING:
    from sentence_transformers import SentenceTransformer


def create_nim_client(base_url, api_key):
    openai_client = OpenAI(base_url=base_url, api_key=api_key)
    return openai_client


def create_hf_model(model_name_or_path: str) -> "SentenceTransformer":
    from sentence_transformers import SentenceTransformer

    return SentenceTransformer(model_name_or_path, trust_remote_code=True)


class HardNegativeMiner:
    """
    Main class that generates annotated training datasets for retriever customization
    This the main class that performs hard negative mining, it takes in
    (questions, documents) tuples as inputs and generates
    (questions, positive documents, negative documents) triplets as outputs

    """

    def __init__(
        self,
        cfg: RetrieverHardNegativeMiningConfig,
    ):

        self.model_name = cfg.model_name
        self.model_type = cfg.model_type
        self.base_url = cfg.base_url
        self.api_key = cfg.api_key
        self.truncate = cfg.truncate
        self.n_hard_negatives = cfg.hard_negatives_to_mine

        if cfg.passage_prefix:
            self.passage_prefix = cfg.passage_prefix
        if cfg.query_prefix:
            self.query_prefix = cfg.query_prefix
        if cfg.hard_neg_mining_algorithm:
            self.hard_neg_mining_algorithm = cfg.hard_neg_mining_algorithm
        else:
            print(
                "hard negative mining algorithm not mentioned in config, using default"
            )
            self.hard_neg_mining_algorithm = "topk_percpos"
        if self.hard_neg_mining_algorithm == "topk_percpos":
            if cfg.percpos:
                self.percpos = cfg.percpos
            else:
                self.percpos = 0.95
        elif self.hard_neg_mining_algorithm == "topk_abs":
            if cfg.max_hardness_threshold:
                self.max_neg_score_threshold = cfg.max_hardness_threshold
            else:
                raise ValueError("Hard negative threshold is required!")
            if cfg.min_hardness_threshold:
                self.min_neg_score_threshold = cfg.min_hardness_threshold
            else:
                self.min_neg_score_threshold = 0.0

        if cfg.min_number_clusters:
            self.min_number_clusters = cfg.min_number_clusters
        if cfg.cluster_output_dir:
            self.cluster_output_dir = cfg.cluster_output_dir
        if cfg.logger_output_dir:
            self.logger_output_dir = cfg.logger_output_dir

    def __dask_tokenize__(self):
        return normalize_token(HardNegativeMiner)

    def assign_ids(self, partition):
        return partition.assign(doc_id=np.arange(len(partition)) + partition.index[0])

    def repartition_semantic_similarity(
        self, dataset: DocumentDataset
    ) -> DocumentDataset:
        df = dataset.df
        n_data = df.shape[0].compute()  # number of row items
        print(f"number of documents in the datasets = {n_data}")
        if self.min_number_clusters >= n_data:
            print("Using too many clusters not recommended!")
            print("Using 1/10th number of datapoints as number of clusters instead.")
            n_clusters = min(1, np.floor(n_data / 10))
        else:
            n_clusters = self.min_number_clusters

        print("Number of clusters used = {}".format(n_clusters))
        assert "doc_id" not in df.columns
        df["embeddings"] = ""  # refers to document embeddings
        df = df.explode("documents")
        df = df.map_partitions(self._get_doc_embeddings, meta=df)
        # df = dd.from_pandas(pdf)
        df = df.map_partitions(self.assign_ids)
        embeddings_dataset = DocumentDataset(df)
        self.clustering_model = ClusteringModel(
            id_column="doc_id",
            max_iter=100,
            n_clusters=n_clusters,
            clustering_output_dir=self.cluster_output_dir,
            logger=self.logger_output_dir,
            keep_all_columns=True,
        )
        clustered_dataset = self.clustering_model(embeddings_dataset)
        df_c = clustered_dataset.df
        df_c = df_c[["documents", "question"]]

        return DocumentDataset(df_c)

    def _get_doc_embeddings(self, p_df: pd.DataFrame):

        if self.model_type == "nvidia":
            self.client = load_object_on_worker(
                attr="nim_embedding_model",
                load_object_function=create_nim_client,
                load_object_kwargs={"base_url": self.base_url, "api_key": self.api_key},
            )

            p_df["embeddings"] = p_df["documents"].map(
                lambda t: self._get_nim_embedding(t, "passage")
            )
        elif self.model_type == "hf":
            self.hf_model = load_object_on_worker(
                attr="hf_embedding_model",
                load_object_function=create_hf_model,
                load_object_kwargs={"model_name_or_path": self.model_name},
            )

            p_df["embeddings"] = p_df["documents"].map(
                lambda t: self._get_hf_embedding(t, self.passage_prefix)
            )
        return p_df

    def _groupby_question(self, pdf):
        pdf2 = pdf.groupby("question").agg({"documents": set})
        pdf2["documents"] = pdf2["documents"].map(lambda x: list(x))
        return pdf2

    def __call__(self, dataset: DocumentDataset) -> DocumentDataset:

        df = dataset.df
        df = df[["question", "documents"]]
        df = df.map_partitions(self._groupby_question).reset_index()
        print("Number partitions in dataset = {}".format(df.npartitions))

        df["neg_doc_scores"] = ""
        df["neg_doc"] = ""
        df["doc_embed"] = ""
        df["query_embed"] = ""
        df["min_pos_score"] = ""

        df = df.map_partitions(self._process_partition, meta=df)

        df = df.rename(columns={"documents": "pos_doc"})
        df = df[["question", "pos_doc", "neg_doc"]]

        return DocumentDataset(df)

    def _process_partition(self, df_p: pd.DataFrame):

        if self.model_type == "nvidia":
            self.client = load_object_on_worker(
                attr="nim_embedding_model",
                load_object_function=create_nim_client,
                load_object_kwargs={"base_url": self.base_url, "api_key": self.api_key},
            )
            df_p["doc_embed"] = df_p["documents"].map(
                lambda pgs: [self._get_nim_embedding(t, "passage") for t in pgs]
            )
            df_p["query_embed"] = df_p["question"].map(
                lambda x: self._get_nim_embedding(x, "query")
            )
        elif self.model_type == "hf":
            self.hf_model = load_object_on_worker(
                attr="hf_embedding_model",
                load_object_function=create_hf_model,
                load_object_kwargs={"model_name_or_path": self.model_name},
            )
            df_p["doc_embed"] = df_p["documents"].map(
                lambda pgs: [
                    self._get_hf_embedding(t, self.passage_prefix) for t in pgs
                ]
            )
            df_p["query_embed"] = df_p["question"].map(
                lambda x: self._get_hf_embedding(x, self.query_prefix)
            )

        doc_embeds = list(itertools.chain(*df_p["doc_embed"].to_list()))
        docs = list(itertools.chain(*df_p["documents"].to_list()))

        if self.hard_neg_mining_algorithm == "topk_abs":
            df_p["neg_doc_scores"] = df_p[["query_embed", "documents"]].apply(
                lambda row: self._get_scores_topk_abs(
                    row["query_embed"], doc_embeds, docs, row["documents"]
                ),
                axis=1,
            )

        elif self.hard_neg_mining_algorithm == "topk_percpos":
            df_p["min_pos_score"] = df_p[["query_embed", "doc_embed"]].apply(
                lambda row: self._get_min_pos_score(row), axis=1
            )
            df_p["neg_doc_scores"] = df_p[["query_embed", "min_pos_score"]].apply(
                lambda row: self._get_scores_topk_percpos(row, doc_embeds, docs), axis=1
            )

        df_p["neg_doc"] = df_p["neg_doc_scores"].map(
            lambda x: [doc for doc, score in x]
        )
        return df_p

    def _get_min_pos_score(self, row):
        x_ = np.array(row["query_embed"])
        y_ = np.array(row["doc_embed"])  # document embeddings
        scores = np.dot(x_, y_.T)
        return np.min(scores)

    def _get_scores_topk_percpos(self, row, docs_embed, docs):
        x_ = np.array(row["query_embed"])
        y_ = np.array(docs_embed)
        scores = np.dot(x_, y_.T)
        neg_docs = []
        neg_docs_scores = []
        max_neg_score_threshold = row["min_pos_score"] * self.percpos
        for idx, s in enumerate(scores):
            if s <= max_neg_score_threshold:
                if docs[idx] not in neg_docs:
                    neg_docs.append(docs[idx])
                    neg_docs_scores.append((docs[idx], s))
        del neg_docs, scores
        return sorted(neg_docs_scores, reverse=True, key=lambda x: x[1])[
            : self.n_hard_negatives
        ]

    def _get_scores_topk_abs(self, x, docs_embed, docs, pos_docs):
        x_ = np.array(x)
        y_ = np.array(docs_embed)
        scores = np.dot(x_, y_.T)
        neg_docs = []
        neg_docs_scores = []
        for idx, s in enumerate(scores):
            if s <= self.max_neg_score_threshold:
                if docs[idx] not in pos_docs:
                    if docs[idx] not in neg_docs:
                        neg_docs.append(docs[idx])
                        neg_docs_scores.append((docs[idx], s))
        del neg_docs, scores
        return sorted(neg_docs_scores, reverse=True, key=lambda x: x[1])[
            : self.n_hard_negatives
        ]

    def _get_hf_embedding(self, text, prefix="query"):
        embeddings = self.hf_model.encode(prefix + text)
        return embeddings

    def _get_nim_embedding(self, text, input_type):
        # Obtain embeddings from nim model
        if isinstance(text, list):
            input_ = text
        elif isinstance(text, str):
            input_ = [text]

        try:
            response = self.client.embeddings.create(
                input=input_,
                model=self.model_name,
                encoding_format="float",
                extra_body={"input_type": input_type, "truncate": self.truncate},
            )
        except Exception as e:
            print(f"Error: {e}")
            response = None

        if response:
            if isinstance(text, list):
                embeddings = [r.embedding for r in response.data]
            elif isinstance(text, str):
                embeddings = response.data[0].embedding
            return embeddings
        else:
            return []



# Copyright (c) 2024, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import nemo_run as run
from nemo_run.core.execution import SlurmExecutor

from nemo_curator.nemo_run import SlurmJobConfig


@run.factory
def nemo_curator_slurm_executor() -> SlurmExecutor:
    """
    Configure the following function with the details of your Slurm cluster
    """
    return SlurmExecutor(
        job_name_prefix="nemo-curator",
        account="my-account",
        nodes=2,
        exclusive=True,
        time="04:00:00",
        container_image="nvcr.io/nvidia/nemo:dev",
        container_mounts=["/path/on/machine:/path/in/container"],
    )


def main():
    # Path to NeMo-Curator/examples/slurm/container_entrypoint.sh on the Slurm cluster
    container_entrypoint = "/cluster/path/slurm/container_entrypoint.sh"
    # The NeMo Curator command to run
    # This command can be susbstituted with any NeMo Curator command
    curator_command = "text_cleaning --input-data-dir=/path/to/data --output-clean-dir=/path/to/output"
    curator_job = SlurmJobConfig(
        job_dir="/home/user/jobs",
        container_entrypoint=container_entrypoint,
        script_command=curator_command,
    )

    executor = run.resolve(SlurmExecutor, "nemo_curator_slurm_executor")
    with run.Experiment("example_nemo_curator_exp", executor=executor) as exp:
        exp.add(curator_job.to_script(), tail_logs=True)
        exp.run(detach=False)


if __name__ == "__main__":
    main()



# Copyright (c) 2024, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import asyncio
import os
import traceback
from typing import List, Optional, Tuple

import numpy as np
import pandas as pd
from tqdm.asyncio import tqdm

from nemo_curator import AsyncLLMClient
from nemo_curator.synthetic import AsyncNemotronGenerator

PROMPT_GENERATE_QUESTIONS_FROM_ANSWER = """TEXT:
{document}

Given the above text, generate exactly {n_openlines} questions that can be answered by the text. All questions must be answerable by the text and be relevant to the text.
Do not directly reference the text in the questions.
Every question should be a complete sentence and end with a question mark. There should be no other text besides the questions.
Begin each question with `* ` and end each question with a newline character. Also, each question must be concise.
Make sure to generate exactly {n_openlines} questions.
"""

PROMPT_PARAPHRASE_TEXT = """TEXT:
{document}

Given the above text, paraphrase the text. Produce exactly {n_openlines} variants.
There should be no other text besides the paraphrased text.
The paraphrased text must be shorter than the original text. The paraphrased text must be factually correct and relevant to the original text.
Begin each variant with `* ` and end each variant with a newline character.
Make sure to generate exactly {n_openlines} variants.
"""


class SyntheticGenerator:

    def __init__(
        self,
        async_llm_client: AsyncLLMClient,
        sdg_model: str,
        sdg_model_kwargs: dict,
        reward_model: Optional[str],
        n_variants: int = 1,
        random_seed: int = 42,
        max_concurrent_entries: int = 320,
    ):
        """
        Initializes the SyntheticGenerator object.

        Args:
            async_llm_client: The asynchronous LLM client.
            sdg_model: The path to the SDG model.
            sdg_model_kwargs: Additional keyword arguments for the SDG model.
            reward_model: The reward model for quality assignment (optional). If not provided,
                the quality will be assigned based on the human-assigned scores.
            n_variants: The number of variants to generate (default: 1).
            random_seed: The random seed for reproducibility (default: 42).
            max_concurrent_entries: The maximum number of concurrent entries (default: 320).
        """
        self.client = async_llm_client
        self.generator = AsyncNemotronGenerator(self.client)
        self.random_state = np.random.RandomState(random_seed)
        self.sdg_model = sdg_model
        self.sdg_model_kwargs = sdg_model_kwargs
        self.reward_model = reward_model
        self.n_variants = n_variants
        self.max_concurrent_entries = max_concurrent_entries

    def run(
        self,
        source_df: pd.DataFrame,
        out_dir: str,
        synth_prefix: str,
        synth_gen_ratio: float,
    ) -> str:
        """
        Runs the synthetic data generation process.

        Args:
            source_df: The source dataframe containing the original data.
            out_dir: The output directory where the synthetic data will be saved.
            synth_prefix: The prefix to be added to the names of the synthetic data files.
            synth_gen_ratio: The ratio of synthetic data to be generated compared to the original data.

        Returns:
            The path to the directory where the synthetic data is saved.
        """
        return asyncio.run(
            self._synthesize_from_source(
                source_df, out_dir, synth_prefix, synth_gen_ratio
            )
        )

    def _split_sdg_responses(self, sdg_response: str) -> List[str]:
        """
        Splits the SDG response into a list of individual entries.

        Args:
            sdg_response: The SDG response string.

        Returns:
            A list of individual SDG entries.
        """
        return [
            entry.strip("*").strip() for entry in sdg_response[0].split("\n") if entry
        ]

    def _write_all_to_file(self, gen_entries, out_fp: str):
        """
        Write all generated synthetic data to a JSON file. If nothing was generated, skip writing to the file.

        Args:
            gen_entries: List of generated entries.
            out_fp: Output file path.
        """
        synth_titles = []
        synth_questions = []
        synth_answers = []
        synth_scores = []
        synth_filenames = []
        synth_ids = []
        synth_tags = []

        for gen_idx, (row_slice, gen_entry) in enumerate(gen_entries):

            for idx, (titles, questions, answers, scores) in enumerate(gen_entry):

                if (
                    len(questions) != self.n_variants
                    or len(answers) != self.n_variants
                    or len(titles) != self.n_variants
                ):
                    print(
                        f"    Skipping synthetic record at ({gen_idx}, {idx}) due to unexpected lengths. The LLM may have failed to generate the expected number of variants."
                    )
                    continue

                row = row_slice.iloc[idx]
                synth_titles.extend(titles)
                synth_questions.extend(questions)
                synth_answers.extend(answers)
                synth_scores.extend(scores)
                synth_filenames.extend([row["file_name"] + ".synth"] * self.n_variants)
                synth_ids.extend(
                    [f"{row['id']}-synth-{i}" for i in range(self.n_variants)]
                )
                synth_tags.extend([row["tags"]] * self.n_variants)

        if not synth_titles:
            print("    No valid synthetic data generated. Skipping writing to file.")
            return

        # Create a Series object with the generated data.
        gen_data = {
            "answer": synth_answers,
            "answer_score": synth_scores,
            "file_name": synth_filenames,
            "id": synth_ids,
            "question": synth_questions,
            "question_score": synth_scores,
            "tags": synth_tags,
            "title": synth_titles,
        }
        # Dump the data into a file.
        gen_df = pd.DataFrame(gen_data)
        gen_df.to_json(out_fp, orient="records", lines=True)

    async def _prompt_model(
        self, row: pd.Series
    ) -> Tuple[List[str], List[str], List[str]]:
        """
        Generates synthetic data by prompting a language model with a given question and answer.

        Args:
            row: The input row containing the question and answer.

        Returns:
            Tuple[List[str], List[str], List[str]]: A tuple containing the generated titles, questions, answers,
            and their corresponding scores.
        """
        question = row["question"]
        answer = row["answer"]

        gen_title = self.generator.generate_closed_qa_instructions(
            document=answer,
            n_openlines=self.n_variants,
            prompt_template=PROMPT_GENERATE_QUESTIONS_FROM_ANSWER,
            model=self.sdg_model,
            model_kwargs=self.sdg_model_kwargs,
        )
        gen_question = self.generator.generate_closed_qa_instructions(
            document=question,
            n_openlines=self.n_variants,
            prompt_template=PROMPT_PARAPHRASE_TEXT,
            model=self.sdg_model,
            model_kwargs=self.sdg_model_kwargs,
        )
        gen_answer = self.generator.generate_closed_qa_instructions(
            document=answer,
            n_openlines=self.n_variants,
            prompt_template=PROMPT_PARAPHRASE_TEXT,
            model=self.sdg_model,
            model_kwargs=self.sdg_model_kwargs,
        )

        gen_title, gen_question, gen_answer = await asyncio.gather(
            gen_title, gen_question, gen_answer
        )

        gen_title = self._split_sdg_responses(gen_title)
        gen_question = self._split_sdg_responses(gen_question)
        gen_answer = self._split_sdg_responses(gen_answer)

        scores = []

        # Use the reward model to assign scores to the generated data.
        if self.reward_model:
            for t, q, a in zip(gen_title, gen_question, gen_answer):
                messages = [
                    {"role": "user", "content": f"{t}\n\n{q}"},
                    {"role": "assistant", "content": a},
                ]
                scores.append(
                    self.client.query_reward_model(
                        messages=messages, model=self.reward_model
                    )
                )

            scores = await asyncio.gather(*scores)
            # Convert each score to a scale of -2 to 2.
            scores = [int(score["helpfulness"] - 2) for score in scores]
        else:
            # Assign a score of 0 to all generated data.
            scores = [0] * self.n_variants

        return gen_title, gen_question, gen_answer, scores

    async def _synthesize_from_source(
        self,
        source_df: pd.DataFrame,
        out_dir_path: str,
        synth_prefix: str,
        synth_gen_ratio: float,
    ) -> str:
        """
        Synthesizes data from a source DataFrame and saves it to a specified directory.

        Args:
            source_df: The source DataFrame containing the data to synthesize.
            out_dir_path: The path to the directory where the synthesized data will be saved.
            synth_prefix: The prefix to use for the synthesized data file.
            synth_gen_ratio: The ratio of data to synthesize from the source DataFrame.

        Returns:
            The path to the directory where the synthesized data is saved.
        """
        os.makedirs(out_dir_path, exist_ok=True)
        # Randomly select a subset of the data to synthesize.
        source_df = source_df.sample(
            frac=synth_gen_ratio, random_state=self.random_state
        )
        prompt_requests = []

        # Generate prompts for each row in the source data and submit them to the LLM.
        for _, row in source_df.iterrows():
            gen_entry = self._prompt_model(row)
            prompt_requests.append(gen_entry)

        gen_entries = []

        for i in tqdm(
            range(0, len(prompt_requests), self.max_concurrent_entries),
            desc=f"Synthesizing {len(source_df)} rows",
        ):
            slice_end = min(i + self.max_concurrent_entries, len(prompt_requests))
            row_slice = source_df[i:slice_end]
            request_slice = prompt_requests[i:slice_end]

            try:
                result = await tqdm.gather(
                    *request_slice, desc=f"---- Rows {i} to {slice_end}"
                )
                gen_entries.append((row_slice, result))
            except Exception as _:
                print(
                    f"    Generation failed for rows {i} to {slice_end} due to the following exception:"
                )
                print("---------------------------------------------------------")
                traceback.print_exc()
                print("---------------------------------------------------------")
                print("Continuing synthetic data generation...")

        # Write the generated data to a file.
        out_fp = f"{out_dir_path}/{synth_prefix}.jsonl"
        self._write_all_to_file(gen_entries, out_fp)
        return out_dir_path



# Copyright (c) 2024, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import csv
import warnings

import sentencepiece
from bs4 import BeautifulSoup
from comment_parser import comment_parser

from nemo_curator.filters.doc_filter import DocumentFilter, import_filter
from nemo_curator.utils.constants import regex_alpha, regex_alphanum
from nemo_curator.utils.text_utils import get_comments_and_docstring


class PythonCommentToCodeFilter(DocumentFilter):

    def __init__(
        self,
        min_comment_to_code_ratio=0.01,
        max_comment_to_code_ratio=0.85,
    ):
        self._min_threshold = min_comment_to_code_ratio
        self._max_threshold = max_comment_to_code_ratio

        self._name = "python_comment_ratio"

    def score_document(self, source):
        docstrings, comments = get_comments_and_docstring(source)
        if docstrings is None or comments is None:
            return 0
        # No need to be super precise about the way this formatted,
        # we just need to count the number of characters.
        return (len(comments) + len(docstrings)) / len(source)

    def keep_document(self, score):
        return self._min_threshold <= score <= self._max_threshold


class GeneralCommentToCodeFilter(DocumentFilter):

    def __init__(
        self,
        language,
        min_comment_to_code_ratio=0.01,
        max_comment_to_code_ratio=0.85,
    ):
        """
        Does not include the comment characters (// or /**/) towards the length of the comment.
        Args:
          language: Mime string of language
        """
        self._lang = language
        self._min_threshold = min_comment_to_code_ratio
        self._max_threshold = max_comment_to_code_ratio

        self._name = "comment_ratio"

    def score_document(self, source):
        try:
            comments = comment_parser.extract_comments_from_str(
                source,
                mime=self._lang,
            )
            comments = " ".join([x.text() for x in comments])
        except Exception:
            warnings.warn("tokenization error, no comment is extracted")
            return 9999
        if comments is None:
            return 0
        return len(comments) / len(source)

    def keep_document(self, score):
        return self._min_threshold <= score <= self._max_threshold


class NumberOfLinesOfCodeFilter(DocumentFilter):

    def __init__(self, min_lines=10, max_lines=20000):
        self._min_lines = min_lines
        self._max_lines = max_lines

        self._name = "num_lines"

    def score_document(self, source):
        return source.count("\n") + 1

    def keep_document(self, score):
        return self._min_lines <= score <= self._max_lines


class TokenizerFertilityFilter(DocumentFilter):

    def __init__(self, path_to_tokenizer=None, min_char_to_token_ratio=2.5):
        if path_to_tokenizer is None:
            raise ValueError(
                "Must provide a valid path to a SentencePiece " "tokenizer"
            )
        self._tokenizer = sentencepiece.SentencePieceProcessor()
        self._tokenizer.Load(path_to_tokenizer)
        self._threshold = min_char_to_token_ratio

        self._name = "tokenizer_fertility"

    def score_document(self, source):
        tokens = self._tokenizer.encode_as_pieces(source)
        num_chars = len(source)
        num_tokens = len(tokens)
        if num_tokens == 0:
            return -1
        return num_chars / num_tokens

    def keep_document(self, score):
        return score >= self._threshold


class XMLHeaderFilter(DocumentFilter):
    """
    This filter tries to identify files that have incorrect file extensions.
    In many cases, these end up being XML files and we try to identify them
    based on the header.
    (Source: Starcoder https://arxiv.org/abs/2305.06161)
    """

    def __init__(self, char_prefix_search_length=100):
        self._char_prefix_search_length = char_prefix_search_length

        self._name = "xml_header"

    def score_document(self, source):
        source_prefix = source[: self._char_prefix_search_length]
        if "<?xml version=" in source_prefix:
            return 1
        else:
            return 0

    def keep_document(self, score):
        return score != 1


class AlphaFilter(DocumentFilter):
    """
    This filter tries to identify files that have large tensors, or tables stored
    as raw text within code files.
    (Source: Starcoder https://arxiv.org/abs/2305.06161)
    """

    def __init__(self, min_alpha_ratio=0.25):
        self._min_alpha_ratio = min_alpha_ratio

        self._name = "alpha_filter"

    def score_document(self, source):
        return len(regex_alpha.findall(source)) / len(source)

    def keep_document(self, score):
        return score >= self._min_alpha_ratio


class HTMLBoilerplateFilter(DocumentFilter):
    """
    This filter tries to identify HTML that is largely boilerplate.
    """

    def __init__(self, min_lang_content_ratio=0.2, min_lang_content_num_chars=100):
        self._min_lang_content_ratio = min_lang_content_ratio
        self._min_lang_content_num_chars = min_lang_content_num_chars

        self._name = "html_boilerplate"

    def score_document(self, source):
        try:
            soup = BeautifulSoup(source, features="html.parser")
        except (TypeError, UnboundLocalError):
            return None

        # kill all script and style elements
        for script in soup(["script", "style"]):
            script.extract()  # rip it out

        # get text
        text = soup.get_text()
        ratio = len(text) / len(source)

        if len(text) < self._min_lang_content_num_chars:
            return 0

        return ratio

    def keep_document(self, score):
        return score >= self._min_lang_content_ratio


class PerExtensionFilter(DocumentFilter):
    """
    This filter that has specific conditions depending on the file extension.
    """

    def __init__(self, lang, extension, metadata_file="code_meta.csv"):
        self._metadata_file = metadata_file
        self._lang = lang
        self._extension = extension
        self._ext_to_filter = self._load_filter_csv(metadata_file, lang)
        self._name = "per_extension_filter"

    def _load_filter_csv(self, path: str, language: str = None):
        """Load csv file that specifies the filter to apply for each (lang, extension).
        TODO: add some tests. Check that filters are correctly set."""
        # (Lang, extension) -> filter_args
        ext_to_filter = {}
        with open(path) as f:
            for row in csv.DictReader(f):
                # Only take the rows corresponding to the language if specified
                if language is None or row["language"] == language:
                    ext_to_filter[(row["language"], row["extension"])] = (
                        self._get_filter_params(row)
                    )
        assert (
            len(ext_to_filter) > 0
        ), f"Did not find filtering params corresponding to language: `{language}` in: {path}"

        return ext_to_filter

    def _get_filter_params(self, row: dict):
        """Extract filter parameters from csv row"""
        include = row["Include"] == "1"
        try:
            line_max = int(row["Long_line_threshold"])
        except ValueError:
            line_max = None
        line_mean = 100 if line_max else None
        try:
            alphanum_frac = float(row["Alphanum_threshold"])
        except ValueError:
            alphanum_frac = None
        try:
            alphabetic_frac = float(row["Alpha filter"])
        except ValueError:
            alphabetic_frac = None
        return include, line_max, line_mean, alphanum_frac, alphabetic_frac

    def _language_format_from_dataset(self, lang: str):
        """Convert: Language field in dataset -> language field in csv file that defines the filters."""
        # TODO: other special cases?
        if lang == "C#":
            return "c-sharp"
        if lang == "F#":
            return "f-sharp"
        return lang.lower().replace(" ", "-")

    def _line_statistics(self, source):
        lengths = [len(x) for x in source.split("\n")]
        max_length = max(lengths)
        mean_length = (len(source) + 1) / len(lengths) - 1

        return max_length, mean_length

    def _alphanum_fraction(self, source):
        return len(regex_alphanum.findall(source)) / len(source)

    def score_document(self, source):
        """Filter files based on line length and % alphanumeric characters.
        The filtering parameters depend on the file extension, given by `ext_to_filter`
        """
        # Get the filter-params we want to use
        # extension `None` is an empty string in the csv
        try:
            (include, line_max, line_mean, alphanum_frac, alphabetic_frac) = (
                self._ext_to_filter[
                    (
                        self._language_format_from_dataset(self._lang),
                        self._extension if self._extension is not None else "",
                    )
                ]
            )
        except KeyError as e:
            # Some extensions are not in the csv. This happens for dockerfiles.
            # Exclude these files
            print(str(e) + f":{self._extension} not in ext_to_filter")
            include = False

        if not include:
            return 0

        max_length, mean_length = self._line_statistics(source)

        if line_max and max_length > line_max:
            return 0

        elif line_mean and mean_length > line_mean:
            return 0

        # Filter files with low percentage of alphanumeric chars
        elif alphanum_frac and self._alphanum_fraction(source) < alphanum_frac:
            return 0

        # Filter files with low percentage of alphabetic chars
        elif alphabetic_frac and sum(map(str.isalpha, source)) < alphabetic_frac * len(
            source
        ):
            return 0

        return 1

    def keep_document(self, score):
        if score is None or score == 0:
            return False
        else:
            return True



# Copyright (c) 2024, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from pathlib import Path

import pytest

from nemo_curator.datasets.parallel_dataset import ParallelDataset


# The source/target file paths will be concatenated as document ID in `ParallelDataset`, so we can't directly pass a `Path` object.
@pytest.fixture
def src_file(pytestconfig):
    root_dir = Path(pytestconfig.rootdir)
    return (root_dir / "tests" / "bitext_data" / "toy.de").absolute().as_posix()


@pytest.fixture
def tgt_file(pytestconfig):
    root_dir = Path(pytestconfig.rootdir)
    return (root_dir / "tests" / "bitext_data" / "toy.en").absolute().as_posix()


class TestReadSimpleBitext:

    def test_pandas_read_simple_bitext(self, src_file, tgt_file):
        ds = ParallelDataset.read_simple_bitext(
            src_input_files=[src_file],
            tgt_input_files=[tgt_file],
            src_lang="de",
            tgt_lang="en",
            backend="pandas",
        )

        for idx, (src_line, tgt_line) in enumerate(
            zip(open(src_file, encoding="utf-8"), open(tgt_file, encoding="utf-8"))
        ):
            assert ds.df["src"].compute()[idx] == src_line.rstrip("\n")
            assert ds.df["tgt"].compute()[idx] == tgt_line.rstrip("\n")
            assert ds.df["src_lang"].compute()[idx] == "de"
            assert ds.df["tgt_lang"].compute()[idx] == "en"

    @pytest.mark.gpu
    def test_cudf_read_simple_bitext(self, src_file, tgt_file):
        ds = ParallelDataset.read_simple_bitext(
            src_input_files=[src_file],
            tgt_input_files=[tgt_file],
            src_lang="de",
            tgt_lang="en",
            backend="cudf",
        )

        for idx, (src_line, tgt_line) in enumerate(
            zip(open(src_file, encoding="utf-8"), open(tgt_file, encoding="utf-8"))
        ):
            assert ds.df["src"].compute()[idx] == src_line.rstrip("\n")
            assert ds.df["tgt"].compute()[idx] == tgt_line.rstrip("\n")
            assert ds.df["src_lang"].compute()[idx] == "de"
            assert ds.df["tgt_lang"].compute()[idx] == "en"
